//
//  game.cpp
//  3ds_metroid
//
//  Created by jbr373 on 28/02/16.
//

#include <3ds.h>
#include <time.h>

#include "game.h"

#define MAX_OUTRO_FRAMES        70

#define INTRO_SECONDS           6

// Globals
u8 stage_object = 0;
u8 game_outro_frames = 0;

bool game_paused = false;
bool game_bgm_reset = false;

time_t intro_time = 0;
time_t intro_time_end;

// Functions

// Special
void gameplay_load_stage(Sound* pSound, SpriteHandler* pSpriteHandler, Stage* pStage, Player* pPlayer);
bool gameplay_play_intro(Sound* pSound, Stage* pStage, Player* pPlayer);
bool gameplay_play_outro(Sound* pSound, Stage* pStage, Player* pPlayer);

// Update
void gameplay_update_projectiles(Sound* pSound, Stage* pStage, Player* pPlayer);
void gameplay_update_items(Sound* pSound, SaveFile* pSaveFile, Stage* pStage, Player* pPlayer);
void gameplay_update_enemies(Sound* pSound, SaveFile* pSaveFile, Physics* pPhysics, Stage* pStage, Player* pPlayer);
void gameplay_update_objects(Sound* pSound, SaveFile* pSaveFile, Stage* pStage, Player* pPlayer);

/********************************************************************************************************************************/
/********************************************************************************************************************************/
/*                                                      GAMEPLAY MAIN                                                           */
/********************************************************************************************************************************/
/********************************************************************************************************************************/

GamePlayStates gameplay(InputInfo& sInputInfo, SpriteHandler* pSpriteHandler, SaveFile* pSaveFile, Sound* pSound, Player* pPlayer, Stage* pStage, Physics* pPhysics, BottomMenu* pBottomMenu)
{
    /*************************************************************************************************************************/
    /*                                                 Player Death                                                          */
    /*************************************************************************************************************************/
    
    if (pPlayer->player_is_dead() == true)
    {
        // Reset Intro
        intro_time = 0;
        game_bgm_reset = false;
        
        // Return Depending on Outro Playing
        if (gameplay_play_outro(pSound, pStage, pPlayer) == true)
        {
            // Outro Event finished playing
            return GAMEPLAYSTATE_GAMEOVER;
        }
        else
        {
            // Outro Event still playing
            return GAMEPLAYSTATE_EVENT;
        }
    }
    
    /************************************************************************************************************************/
    /*                                                 Game Paused                                                          */
    /************************************************************************************************************************/
    
    // Check Pause Button
    if (sInputInfo.KeyDown & KEY_START)
    {
        if (game_paused == false)
        {
            game_paused = true;
        }
        else
        {
            game_paused = false;
        }
    }
    
    /************************************************************************************************************************/
    /*                                                  Load Stage                                                          */
    /************************************************************************************************************************/
    
    // Load Stage Data
    if (pStage->stage_get_room_change() == STAGE_ROOMCHANGE_LOAD)
    {
        // Stop Shooting / Running Sound
        pSound->sound_stop(SOUNDTYPE_PLAYER_SHOOTING);
        pSound->sound_stop(SOUNDTYPE_PLAYER_MOVING);
        
        // Reset Ledge States
        if (pPlayer->player_get_state() == PLAYERSTATE_CLIMBING || pPlayer->player_get_state() == PLAYERSTATE_LEDGE)
        {
            // Set State
            pPlayer->player_set_state(PLAYERSTATE_IDLE);
            
            // Set Animation
            pPlayer->player_set_anim(PLAYERANIM_STANDING);
        }
        
        // Load Next Stage
        gameplay_load_stage(pSound, pSpriteHandler, pStage, pPlayer);
        
        // Load Map File if necessary
        if (pStage->stage_get_stage_phase() != pSaveFile->savefile_get_stage_phase())
        {
            // Set Stage Phase
            pSaveFile->savefile_set_stage_phase(pStage->stage_get_stage_phase());
            
            // Load Minimap File
            pBottomMenu->gamemenu_reload_minimap(pStage->stage_get_stage_phase());
        }
        
        // Check if our Room Flag is already set
        if (pSaveFile->savefile_get_minimap_flag() & pStage->stage_get_minimap_id())
        {
            // Already set, do nothing
        }
        else
        {
            // Add new Room Flag
            pSaveFile->savefile_set_minimap_flag(pStage->stage_get_minimap_id());
        }
        
        // Set New Minimap Position
        if (pBottomMenu->gamemenu_get_minimap() != NULL)
        {
            if (pBottomMenu->gamemenu_get_minimap()->minimapfile_loaded == true && pBottomMenu->gamemenu_get_minimap_follow_auto() == false)
            {
                // Shift Minimap to center the next Room
                pBottomMenu->gamemenu_force_minimap_shift(pStage->stage_get_minimap_id());
            }
        }
    }
    
    /************************************************************************************************************************/
    /*                                          Load BGM After Intro Sound                                                  */
    /************************************************************************************************************************/
    
    // Play Intro Event
    if (gameplay_play_intro(pSound, pStage, pPlayer) == true)
    {
        return GAMEPLAYSTATE_EVENT;
    }
    
    /************************************************************************************************************************/
    /*                                          Return in Specific Cases                                                    */
    /************************************************************************************************************************/
    
    // Return on Room Fade-In or Fade-Out
    if (pStage->stage_get_room_change() != STAGE_ROOMCHANGE_NOCHANGE)
    {
        return GAMEPLAYSTATE_RUNNING;
    }

    // Return on Game Pause
    if (game_paused == true)
    {
        return GAMEPLAYSTATE_PAUSED;
    }
    
    /************************************************************************************************************************/
    /*                                                  Get Collision                                                       */
    /************************************************************************************************************************/
    
    // Player Collision
    u8 collision = 0;
    
    collision = pStage->stage_check_collision(pPlayer->player_get_x(), pPlayer->player_get_y(), pPlayer->player_get_width(), pPlayer->player_get_height());
 
    /************************************************************************************************************************/
    /*                                              Get Player Input                                                        */
    /************************************************************************************************************************/
    
    // Player Update
    pPlayer->player_action(pSaveFile, pSound, sInputInfo.KeyDown, sInputInfo.KeyHeld, sInputInfo.KeyUp, collision,
                           pStage->stage_is_in_water_steps(pPlayer->player_get_x(), pPlayer->player_get_y(), pPlayer->player_get_height()));
    
    /************************************************************************************************************************/
    /*                                               Player Physics                                                         */
    /************************************************************************************************************************/
    
    // Update Physics
    pPhysics->physics_player_physics(pPlayer, pStage, pSaveFile, collision);
    
    /************************************************************************************************************************/
    /*                                                Projectiles                                                           */
    /************************************************************************************************************************/
    
    // Update Player Projectiles
    gameplay_update_projectiles(pSound, pStage, pPlayer);
    
    /************************************************************************************************************************/
    /*                                                   Actors                                                             */
    /************************************************************************************************************************/
    
    // Update Enemies
    if (pStage->stage_get_enemy_count() > 0)
    {
        gameplay_update_enemies(pSound, pSaveFile, pPhysics, pStage, pPlayer);
    }
    
    // Update Items
    if (pStage->stage_get_items_count() > 0)
    {
        gameplay_update_items(pSound, pSaveFile, pStage, pPlayer);
    }
    
    // Update Objects
    if (pStage->stage_get_object_count() > 0)
    {
        gameplay_update_objects(pSound, pSaveFile, pStage, pPlayer);
    }

    /************************************************************************************************************************/
    /*                                                  Bottom Screen Menu                                                  */
    /************************************************************************************************************************/
    
    switch (pBottomMenu->gamemenu_action(sInputInfo.CurrentTouchPosition))
    {
        case BOTTOMMENU_SETMAP:
            pBottomMenu->gamemenu_set_submenu(SUBMENU_MAP);
            break;
        case BOTTOMMENU_SETSTATUS:
            pBottomMenu->gamemenu_set_submenu(SUBMENU_STATUS);
            break;
        case BOTTOMMENU_ICEBEAM:
            if (pSaveFile->savefile_get_icebeam() == true)
                pSaveFile->savefile_set_activebeam(BEAM_ICE);
            break;
        case BOTTOMMENU_WAVEBEAM:
            if (pSaveFile->savefile_get_wavebeam() == true)
                pSaveFile->savefile_set_activebeam(BEAM_WAVE);
            break;
        case BOTTOMMENU_SPAZERBEAM:
            if (pSaveFile->savefile_get_spazerbeam() == true)
                pSaveFile->savefile_set_activebeam(BEAM_SPAZER);
            break;
        case BOTTOMMENU_PLASMABEAM:
            if (pSaveFile->savefile_get_plasmabeam() == true)
                pSaveFile->savefile_set_activebeam(BEAM_PLASMA);
            break;
            
        default:
            pBottomMenu->gamemenu_set_submenu(SUBMENU_MAP);
            break;
    }
    
    // Set Bottom Screen Selection Coordinates
    pBottomMenu->gamemenu_set_sel_pos(BOTTOMMENU_SETSTATUS + pSaveFile->savefile_get_activebeam());
    
    // Return Gameplay Running
    return GAMEPLAYSTATE_RUNNING;
}

/********************************************************************************************************************************/
/********************************************************************************************************************************/
/*                                                    GAMEPLAY ELEMENTS                                                         */
/********************************************************************************************************************************/
/********************************************************************************************************************************/

/********************************************************************************************************************************/
/*                                                      GamePlay Intro                                                          */
/********************************************************************************************************************************/

bool gameplay_play_intro(Sound* pSound, Stage* pStage, Player* pPlayer)
{
    // Return on already executed
    if (game_bgm_reset == true)
    {
        return false;
    }
    
    bool result = true;
    
    if (intro_time == 0)
    {
        // Set Up Time
        intro_time = time(NULL);
        intro_time_end = time(NULL) + INTRO_SECONDS;
    }
    else
    {
        // Update Time or Play Intro
        if (intro_time < intro_time_end)
        {
            // Update Time
            intro_time = time(NULL);
        }
        else
        {
            // Reset Outro Frames
            game_outro_frames = 0;
            
            // Play BGM
            pSound->sound_play_bgm(pStage->stage_get_sound_path(), true, true);
            
            // Set Animation
            pPlayer->player_set_anim(PLAYERANIM_STANDING);
            
            // Flag BGM
            game_bgm_reset = true;
            
            // Update Result
            result = false;
        }
    }
    
    // Return Result
    return result;
}

/********************************************************************************************************************************/
/*                                                      GamePlay Outro                                                          */
/********************************************************************************************************************************/

bool gameplay_play_outro(Sound* pSound, Stage* pStage, Player* pPlayer)
{
    // Force Animation
    pPlayer->player_set_anim(PLAYERANIM_DEATH);
    
    // Set Player X
    if (pPlayer->player_get_x() < 200)
    {
        pPlayer->player_set_x(pPlayer->player_get_x() + 6);
    }
    else if (pPlayer->player_get_x() > pStage->stage_get_width() - 200)
    {
        pPlayer->player_set_x(pPlayer->player_get_x() - 6);
    }
    
    
    // Set Player Y
    if (pPlayer->player_get_y() < 120)
    {
        pPlayer->player_set_y(pPlayer->player_get_y() + 6);
    }
    else if (pPlayer->player_get_y() > pStage->stage_get_height() - 120)
    {
        pPlayer->player_set_y(pPlayer->player_get_y() - 6);
    }

    // Outro Frame Counter
    if (game_outro_frames >= MAX_OUTRO_FRAMES)
    {
        // Set Animation
        if (pStage->stage_get_room_change() == STAGE_ROOMCHANGE_LOAD)
        {
            // Return Max Outro Frames reached
            return true;
        }
        else if (pStage->stage_get_room_change() != STAGE_ROOMCHANGE_FADEOUT)
        {
            // Set Fade Out
            pStage->stage_set_room_change(STAGE_ROOMCHANGE_FADEOUT);
            
            // Fade-Out Still Playing
            return false;
        }
        else
        {
            // Wrong animation, installation finished
            return false;
        }
    }
    else
    {
        // Play Death Sound
        if (game_outro_frames == 1)
        {
            pSound->sound_play_fx(SOUND_PLAYER_DEATH, SOUNDTYPE_PLAYER_MOVING, false, true);
        }
        
        // Increment Outro Frames
        ++game_outro_frames;
        
        // Outro Still Playing
        return false;
    }
}

/*******************************************************************************************************************************/
/*                                                      Stage Loading                                                          */
/*******************************************************************************************************************************/

void gameplay_load_stage(Sound* pSound, SpriteHandler* pSpriteHandler, Stage* pStage, Player* pPlayer)
{
    // Set Player Position
    pPlayer->player_set_x(pStage->stage_get_object(stage_object)->object_get_value_b());
    pPlayer->player_set_y(pStage->stage_get_object(stage_object)->object_get_value_c());
    
    // Delete Stage-Actors specific Graphics
    pSpriteHandler->spritehandler_clear_all_sprites();
    
    // Delete Old Stage-specific Objects
    pPlayer->player_delete_all_projectiles();
        
    // Load Stage Elements
    pStage->stage_load(pSound, pSpriteHandler, pStage->stage_get_object(stage_object)->object_get_value_a());
  
    // Set Animation
    pStage->stage_set_room_change(STAGE_ROOMCHANGE_FADEIN);
}

/*****************************************************************************************************************************/
/*                                                      Projectiles                                                          */
/*****************************************************************************************************************************/

void gameplay_update_projectiles(Sound* pSound, Stage* pStage, Player* pPlayer)
{
    // Set Collision Variable
    u8 collision_projectile = 0;
    
    // Update Player Projectiles ( Vector Size 1 = Object at Pos 0)
    for (u8 i = 0; i <= pPlayer->player_get_projectile_count(); ++i)
    {
        // Only continue if the current slot is actually used
        if (pPlayer->player_get_projectile(i) != NULL)
        {
            // Check Enemy Hit
            for (u8 j = 0; j <= pStage->stage_get_enemy_count() - 1; ++j)
            {
                // Check for Projectile Hit
                if (pStage->stage_get_enemy(j)->enemy_check_hit_collision(pPlayer->player_get_projectile(i)->projectile_get_x(),
                                                                          pPlayer->player_get_projectile(i)->projectile_get_y()) == true &&
                    pStage->stage_get_enemy(j)->enemy_is_dead() == false)
                {
                    // Damage Enemy
                    pStage->stage_get_enemy(j)->enemy_damage(pPlayer->player_get_projectile(i)->projectile_get_damage(),
                                                             pPlayer->player_get_projectile(i)->projectile_get_type());
                    
                    // Set Enemy Explosion Start
                    pPlayer->player_get_projectile(i)->projectile_set_exploded(true);
                }
            }
            
            // Check Object Hit
            if (pStage->stage_get_object_count() > 0)
            {
                for (u8 k = 0; k <= pStage->stage_get_object_count() - 1; ++k)
                {
                    // Set Hit Type
                    u8 hit_type = OBJECTATTACKTYPE_MISS;
                    
                    switch (pPlayer->player_get_projectile(i)->projectile_get_type())
                    {
                        case PROJECTILE_MISSLE:
                            hit_type = OBJECTATTACKTYPE_MISSLE;
                            break;
                            
                        default:
                            hit_type = OBJECTATTACKTYPE_BEAM;
                            break;
                    }
                    
                    // Only Perform
                    if (pStage->stage_get_object(k)->object_is_destroy_completed() == false)
                    {
                        // Did we collide with the Object?
                        if (pStage->stage_get_object(k)->object_is_collision(pPlayer->player_get_projectile(i)->projectile_get_x(),
                                                                             pPlayer->player_get_projectile(i)->projectile_get_y(),
                                                                             pPlayer->player_get_projectile(i)->projectile_get_hitbox_width(),
                                                                             pPlayer->player_get_projectile(i)->projectile_get_hitbox_height(),
                                                                             pPlayer->player_get_projectile(i)->projectile_get_facing(),
                                                                             true,
                                                                             hit_type) == true)
                        {
                            // Set Enemy Explosion Start
                            pPlayer->player_get_projectile(i)->projectile_set_exploded(true);
                        }
                    }
                }
            }
            
            // Get Collision
            collision_projectile = pStage->stage_check_collision(pPlayer->player_get_projectile(i)->projectile_get_x(),
                                                                 pPlayer->player_get_projectile(i)->projectile_get_y(),
                                                                 pPlayer->player_get_projectile(i)->projectile_get_hitbox_width(),
                                                                 pPlayer->player_get_projectile(i)->projectile_get_hitbox_height());
            
            // Update
            pPlayer->player_get_projectile(i)->projectile_update(pSound, collision_projectile);
            
            // Check for Deletion
            if (pPlayer->player_get_projectile(i)->projectile_get_explode_finished() == true ||
                pPlayer->player_get_projectile(i)->projectile_is_outside_screen(pPlayer->player_get_x(), pPlayer->player_get_y()) == true)
            {
                // Delete Projectile
                pPlayer->player_delete_projectile(i);
            }
        }
    }
}

/***********************************************************************************************************************/
/*                                                      Items                                                          */
/***********************************************************************************************************************/

void gameplay_update_items(Sound* pSound, SaveFile* pSaveFile, Stage* pStage, Player* pPlayer)
{
    // Do for each Item
    for (u8 j = 0; j <= pStage->stage_get_items_count() - 1; ++j)
    {
        // Check if we're in collision with the Item and if it is collected already!
        if (pStage->stage_get_item(j) != NULL &&
            pStage->stage_get_item(j)->item_get_collected() == false &&
            pStage->stage_get_item(j)->item_check_collection(pSaveFile,
                                                             pSound,
                                                             pPlayer->player_get_x(), pPlayer->player_get_y(),
                                                             pPlayer->player_get_height(), pPlayer->player_get_width()))
        {
            // Get Item Actions
            switch (pStage->stage_get_item(j)->item_get_action())
            {
                case ITEMACTION_ADDHEALTH:
                    pPlayer->player_heal(pStage->stage_get_item(j)->item_get_action_value());
                    break;
                case ITEMACTION_ADDMISSLES:
                    pPlayer->player_increase_available_missles(pStage->stage_get_item(j)->item_get_action_value());
                    break;
                case ITEMACTION_INCREASEHEALTHTANKS:
                    pPlayer->player_increase_healthtanks();
                    break;
                case ITEMACTION_INCREASEMISSLETANKS:
                    pPlayer->player_increase_total_missles(pStage->stage_get_item(j)->item_get_action_value());
                    break;
            }
            
            // Set Item Collected
            pStage->stage_get_item(j)->item_set_collected(true);
        }
        else if (pStage->stage_get_item(j) != NULL && pStage->stage_get_item(j)->item_get_collected() == true)
        {
            // Delete Collected Item
            //pStage->stage_delete_item(j);
        }
    }
}

/*************************************************************************************************************************/
/*                                                      Enemies                                                          */
/*************************************************************************************************************************/

void gameplay_update_enemies(Sound* pSound, SaveFile* pSaveFile, Physics* pPhysics, Stage* pStage, Player* pPlayer)
{
    // Collision Value
    u8 collision_enemy = COLLISION_NONE;
    
    // Update Enemies
    for (u8 j = 0; j <= pStage->stage_get_enemy_count() - 1; ++j)
    {
        // Get Enemy Collision
        collision_enemy = pStage->stage_check_collision(pStage->stage_get_enemy(j)->enemy_get_x(),
                                                        pStage->stage_get_enemy(j)->enemy_get_y(),
                                                        pStage->stage_get_enemy(j)->enemy_get_hitbox_width(),
                                                        pStage->stage_get_enemy(j)->enemy_get_hitbox_height());
        // Update Enemy Action
        pStage->stage_get_enemy(j)->enemy_action(pSound, pPlayer, collision_enemy);
        
        // Update Physics
        pPhysics->physics_enemy_physics(pStage->stage_get_enemy(j),
                                        pStage,
                                        pStage->stage_check_collision(pStage->stage_get_enemy(j)->enemy_get_x(),
                                                                      pStage->stage_get_enemy(j)->enemy_get_y(),
                                                                      pStage->stage_get_enemy(j)->enemy_get_hitbox_width(),
                                                                      pStage->stage_get_enemy(j)->enemy_get_hitbox_height())
                                        );
        
        // Check for Enemy Touch Hit
        if (pPlayer->player_get_enemy_collision(pStage->stage_get_enemy(j)->enemy_get_x() + pStage->stage_get_enemy(j)->enemy_get_damage_hitbox_x(),
                                                pStage->stage_get_enemy(j)->enemy_get_y() + pStage->stage_get_enemy(j)->enemy_get_damage_hitbox_y(),
                                                pStage->stage_get_enemy(j)->enemy_get_damage_hitbox_width(),
                                                pStage->stage_get_enemy(j)->enemy_get_damage_hitbox_height()) == true &&
            pStage->stage_get_enemy(j)->enemy_is_dead() == false)
        {
            // Damage Player
            pPlayer->player_damage(pStage->stage_get_enemy(j)->enemy_get_touch_damage());
        }
        
        /*
         // Check Enemy Projectile Hit
         if ()
         {
         // Enemy Projectile hit Player
         }
         */
        
        // Check for Enemy Loot
        if (pStage->stage_get_enemy(j)->enemy_is_dead() == true)
        {
            // Roll Loot
            if (pStage->stage_get_enemy(j)->enemy_drop_loot() == true && strcmp(pStage->stage_get_enemy(j)->enemy_get_loot().c_str(), LINE_DO_NOT_USE) != 0)
            {
                // Drop Loot
                pStage->stage_create_item(pStage->stage_get_enemy(j)->enemy_get_loot(),
                                          pStage->stage_get_enemy(j)->enemy_get_x() + pStage->stage_get_enemy(j)->enemy_get_loot_spawn_x(),
                                          pStage->stage_get_enemy(j)->enemy_get_y() + pStage->stage_get_enemy(j)->enemy_get_loot_spawn_y());
            }
        }
    }
}

/*************************************************************************************************************************/
/*                                                      Objects                                                          */
/*************************************************************************************************************************/

void gameplay_update_objects(Sound* pSound, SaveFile* pSaveFile, Stage* pStage, Player* pPlayer)
{
    // Update Objects with Loop
    for (u8 k = 0; k <= pStage->stage_get_object_count() - 1; ++k)
    {
        // Only Perform
        if (pStage->stage_get_object(k)->object_is_destroy_completed() == false)
        {
            // Check Object Action
            pStage->stage_get_object(k)->object_action(pPlayer->player_get_x(),
                                                       pPlayer->player_get_y(),
                                                       1 /*pPlayer->player_get_hitbox_width()*/,
                                                       1 /*pPlayer->player_get_hitbox_height()*/,
                                                       pPlayer->player_get_facing(),
                                                       pStage->stage_get_width(), pStage->stage_get_height());
            
            // Did we collide with the Object?
            if (pStage->stage_get_object(k)->object_is_collision(pPlayer->player_get_x(),
                                                                 pPlayer->player_get_y(),
                                                                 1 /*pPlayer->player_get_hitbox_width()*/,
                                                                 1 /*pPlayer->player_get_hitbox_height()*/,
                                                                 pPlayer->player_get_facing()))
            {
                switch (pStage->stage_get_object(k)->object_get_type())
                {
                    case OBJECTTYPE_ROOMTRIGGER:
                        stage_object = k;
                        pStage->stage_set_room_change(STAGE_ROOMCHANGE_FADEOUT);
                        break;
                    case OBJECTTYPE_DAMAGE:
                        pPlayer->player_damage(pStage->stage_get_object(k)->object_get_value_a()); // Who to damage
                        pPlayer->player_damage(pStage->stage_get_object(k)->object_get_value_b()); // Actual Damage
                        break;
                    case OBJECTTYPE_SAVESTATION:
                        if (pStage->stage_get_object(k)->object_get_value_a() == 1)
                            pPlayer->player_heal_full();
                        
                        pPlayer->player_save(pSaveFile);
                        break;
                }
            }
        }
    }
}
