//
//  main.cpp
//  3ds_metroid
//
//  Created by jbr373 on 15/01/16.
//

#include <iostream>
#include <3ds.h>
#include <sf2d.h>
#include <sfil.h>

#include "stage.h"
#include "render.h"
#include "image.h"
#include "savefile.h"
#include "filepaths.h"
#include "menu.h"
#include "physics.h"
#include "menu_game.h"
#include "sound.h"
#include "system.h"
#include "sprite_handler.h"
#include "game.h"

#define NDSP_BUFFERS                    5

enum GameStates
{
    GAMESTATE_INIT = 0,
    GAMESTATE_MAINMENU = 1,
    GAMESTATE_LOAD_CONTINUE = 2,
    GAMESTATE_LOAD_SPRITES = 3,
    GAMESTATE_LOAD_GAMEPLAY = 4,
    GAMESTATE_GAMEPLAY = 5,
    GAMESTATE_LOAD_GAMEOVER = 6,
    GAMESTATE_GAMEOVER = 7
};

bool game_debug = false;
extern bool game_paused;

int main(/*int argc, const char * argv[]*/)
{
    /* Initialize */
    
    // Srv
    srvInit();
    
    // Apt
    aptInit();
    
    // Sound
    ndspInit();
    ndspSetOutputMode(NDSP_OUTPUT_STEREO);
    ndspSetOutputCount(NDSP_BUFFERS);
    
    // SF2D + SFIL
    sf2d_init();//_advanced(METROID_GPUCMD_SIZE, METROID_TEMPPOOL_SIZE);
    sf2d_set_3D(0);
    sf2d_set_clear_color(RGBA8(0x00, 0x00, 0x00, 0x00));

    // Classes
    Render* pRender = new Render();
    
    SaveFile* pSaveFile = new SaveFile();
    pSaveFile->savefile_load(game_debug);
    SpriteHandler* pSpriteHandler = NULL;
    Player* pPlayer = NULL;
    Stage* pStage = new Stage();
    GameMenu* pMenu = new GameMenu();
    Physics* pPhysics = new Physics();
    Sound* pSound = new Sound();
    BottomMenu* pGameMenu = new BottomMenu();
    
    // Variables
    InputInfo sInputInfo;
    u8 mainmenu_result = 0;
    bool game_crash = false;
    GameStates game_state = GAMESTATE_INIT;
    GamePlayStates gameplay_state = GAMEPLAYSTATE_RUNNING;
  
    /* Game Main Loop */
    
    // Main Loop
    while (aptMainLoop())
    {
        if (game_debug == true)
        {
            printf("\x1b[26;0H [  Free Linear Space: %ld ]", linearSpaceFree());
            printf("\x1b[27;0H [  Free VRAM Space: %ld ]", vramSpaceFree());
            printf("\x1b[28;0H [  Free Temp Space: %ld ]", sf2d_pool_space_free());
            printf("\x1b[29;0H [  FPS: %ld ]", sf2d_get_fps());
        }
        
        // Get Input
        hidScanInput();
        sInputInfo.KeyDown = hidKeysDown();
        sInputInfo.KeyHeld = hidKeysHeld();
        sInputInfo.KeyUp = hidKeysUp();
        hidTouchRead(&sInputInfo.CurrentTouchPosition);
        
        /* Game States */
        
        if (game_state == GAMESTATE_INIT)
        {
            // Load Sounds
            pSound->sound_load_all();
            
            // Load Main Data
            if (game_debug == true)
            {
                consoleInit(GFX_BOTTOM, NULL);
                consoleClear();
                game_state = GAMESTATE_LOAD_GAMEPLAY;
            }
            else
            {
                game_state = GAMESTATE_MAINMENU;
            }
            
        }
        else if (game_state == GAMESTATE_MAINMENU)
        {
            // Check Main Menu
            mainmenu_result = pMenu->gamemenu_main(pSaveFile, pSound, sInputInfo.KeyDown, sInputInfo.CurrentTouchPosition);
            
            // Action from Main Menu
            if (mainmenu_result == GAMEMENU_QUIT)
            {
                // Exit Main Loop
                break;
            }
            else if (mainmenu_result == GAMEMENU_START)
            {
                // Load Gameplay
                game_state = GAMESTATE_LOAD_GAMEPLAY;
            }
            
            // Draw Main Menu
            pRender->render_draw(pMenu, pSaveFile);
        }
        else if (game_state == GAMESTATE_LOAD_GAMEPLAY || game_state == GAMESTATE_LOAD_CONTINUE)
        {
            // Change Action depending on Gameplay loading type
            if (game_state == GAMESTATE_LOAD_GAMEPLAY)
            {
                // Load Player
                pPlayer = new Player(pSaveFile);
                
                // Create Sprite Handler
                pSpriteHandler = new SpriteHandler(pSaveFile);
            }
            else
            {
                // Reload Save File
                pSaveFile->savefile_load(game_debug);
                
                // Reload Player
                pPlayer->player_reload(pSaveFile);
            }
            
            // Set Controls Settings
            pPlayer->player_set_controls(pMenu->gamemenu_get_config_file()->config_movement_changed,
                                         pMenu->gamemenu_get_config_file()->config_shooting_changed);
            
            // Set Map Settings
            pGameMenu->gamemenu_set_minimap_follow_auto(pMenu->gamemenu_get_config_file()->config_map);

            // Load Stage
            pStage->stage_load(NULL, pSpriteHandler, pSaveFile->savefile_get_room());
            
            // Set Minimap Flag
            if (game_state == GAMESTATE_LOAD_GAMEPLAY)
            {
                // Set Minimap Flag
                pSaveFile->savefile_set_minimap_flag(pStage->stage_get_minimap_id());
            }
            
            // Load Minimap
            pGameMenu->gamemenu_reload_minimap(pSaveFile->savefile_get_stage_phase());

            // Set Animation
            pStage->stage_set_room_change(STAGE_ROOMCHANGE_FADEIN);
            
            // Load Intro Sounds
            pSound->sound_play_bgm(SND_BGM_SAMUS, false, true);

            // Set new Game State
            game_state = GAMESTATE_GAMEPLAY;
        }
        else if (game_state == GAMESTATE_GAMEPLAY)
        {
            // Check Gameplay
            gameplay_state = gameplay(sInputInfo, pSpriteHandler, pSaveFile, pSound, pPlayer, pStage, pPhysics, pGameMenu);
            
            // Kill Game Pause Screen Input
            if (sInputInfo.KeyDown & KEY_SELECT && gameplay_state == GAMEPLAYSTATE_PAUSED)
            {
                break;
            }
            
            // Set Game Over
            if (gameplay_state == GAMEPLAYSTATE_GAMEOVER)
            {
                game_state = GAMESTATE_LOAD_GAMEOVER;
            }
            else
            {
                // Draw Gameplay
                pRender->render_draw(pSpriteHandler, pStage, pPlayer, pGameMenu, pSaveFile, game_paused);
            }
        }
        else if (game_state == GAMESTATE_LOAD_GAMEOVER)
        {
            // Update Main Menu
            pMenu->gamemenu_set_submenu(GAMEMENU_TRANSITION);
            pMenu->gamemenu_set_next_submenu(GAMEMENU_GAMEOVER);
            
            // Delete Stage-Actors specific Graphics
            pSpriteHandler->spritehandler_clear_all_sprites();
            
            // Delete Projectiles
            pPlayer->player_delete_all_projectiles();
            
            // Clean Up Stage
            pStage->stage_delete();
            pStage->stage_set_sound_path("NONE");
            pStage->stage_set_tileset_path("NONE");
            pStage->stage_set_background_path("NONE");
            
            // Play Game Over BGM
            pSound->sound_stop(SOUNDTYPE_BGM);

            // Set Game Over State
            game_state = GAMESTATE_GAMEOVER;
        }
        else if (game_state == GAMESTATE_GAMEOVER)
        {
            // Check Game Over
            mainmenu_result = pMenu->gamemenu_gameover(pSound, sInputInfo.KeyDown, sInputInfo.CurrentTouchPosition);
            
            // Set Action depending on State
            if (mainmenu_result == GAMEMENU_START)
            {
                // Load Gameplay again
                game_state = GAMESTATE_LOAD_CONTINUE;
            }
            else if (mainmenu_result == GAMEMENU_QUIT)
            {
                // Exit Main Loop
                break;
            }
            
            // Draw Main Menu
            pRender->render_draw(pMenu, pSaveFile);
        }
        else
            exit(1);

        // Crash Game if we encounter something bad
        if (game_crash == true)
            exit(1);
        
        // Swap Buffers
        sf2d_swapbuffers();
    }
    
    // Delete Classes
    /*
    delete pRender;
    delete pSound;
    delete pMenu;
    delete pGameMenu;
    delete pSpriteHandler;
    delete pPlayer;
    delete pStage;
    delete pPhysics;
    delete pSaveFile;
    */
    
    // Exit Services
    sf2d_fini();
    ndspExit();
    aptExit();
    srvExit();
    
    // End
    return 0;
}
