//
//  player.cpp
//  3ds_metroid
//
//  Created by jbr373 on 16/01/16.
//

#include <sfil.h>

#include "player.h"
#include "filepaths.h"
#include "stage_collision.h"
#include "actor_facing.h"
#include "render_poscorrection.h"

// Hitboxes

// Standing Hitbox
#define PLAYER_HITBOX_STANDING_W                40
#define PLAYER_HITBOX_STANDING_H                72

// Crouched Hitbox
#define PLAYER_HITBOX_CROUCHED_W                45
#define PLAYER_HITBOX_CROUCHED_H                50

// Morphball Hitbox
#define PLAYER_HITBOX_MORPHBALL_W               30
#define PLAYER_HITBOX_MORPHBALL_H               30

// Position Correction
#define PROJECTILE_HEIGHT_HALF_STANDING         27//30
#define PROJECTILE_HEIGHT_HALF_CROUCHED         32
#define PROJECTILE_AIM_UP_CORRECTION            12//10

// Default Player Variables
#define DEFAULT_LOW_HEALTH                      20
#define DEFAULT_FULL_HEALTH                     99
#define DEFAULT_HEALTHTANK_SIZE                 100
#define DEFAULT_MAX_MISSLES_TOTAL               250
#define DEFAULT_PLAYER_SPEED                    4

// Non-Beam Projectiles
#define USEDPROJECTILE_MISSLE                   98
#define USEDPROJECTILE_BOMB                     99

// Frame Times
#define DEFAULT_LOW_HEALTH_FRAMES               13
#define CHARGEBEAM_PLAYER_SHOOTING_LOCK         2
#define DEFAULT_PLAYER_SHOOTING_LOCK            8
#define DEFAULT_INVINCIBILITY_FRAMES            40
#define DEFAULT_DAMAGE_FRAMES                   12
#define DEFAULT_LEDGE_BLOCK_FRAMES              8

/*
 *      TODO
 * ---------------
 * - std::vector for Projectiles, current solution is performance-unfriendly
 */

Player::Player(SaveFile* pSaveFile)
{
    // Controls
    movement_up = KEY_CPAD_UP;
    movement_down = KEY_CPAD_DOWN;
    movement_left = KEY_CPAD_LEFT;
    movement_right = KEY_CPAD_RIGHT;
    
    beam_up = KEY_DUP;
    beam_down = KEY_DDOWN;
    beam_left = KEY_DLEFT;
    beam_right = KEY_DRIGHT;
    
    shooting_button = KEY_B;
    
    // Player States
    player_anim = PLAYERANIM_FRONT;
    player_facing = FACING_RIGHT;
    player_shoot_facing = FACING_RIGHT;
    player_speed = DEFAULT_PLAYER_SPEED;
    player_state = PLAYERSTATE_IDLE;
    player_collision = COLLISION_NONE;
    player_invincibility_frames = 0;
    player_shooting_lock_frames = 0;
    player_ledge_grab_block_frames = 0;
    player_low_health_counter = 0;
    player_damage_frames = 0;
    player_charge_beam = 0;
    player_last_projectile = 0;
    
    player_x = 0;
    player_y = 0;
    player_width = 1;
    player_height = 1;
    
    player_hitbox_x = 0;
    player_hitbox_width = 1;
    player_hitbox_height = 1;
    
    player_dead = false;
    player_took_damage = false;
    player_shooting = false;
    player_morphball = false;
    player_morphball_running_jump = false;
    player_jump_movement_cancel = false;
    player_crouched = false;
    player_spiderball = false;
    player_barriersuit = false;
    
    // Player Stats
    player_health = 99;
    player_healthtanks_total = 0;
    player_healthtanks_available = 0;
    player_missles_total = 30;
    player_missles_available = 30;
    player_metroids = 39;
    player_phase = 0;
    
    // Player Projectiles
    player_projectiles_count = 0;
    
    for (u8 i = 0; i < PLAYER_MAX_PROJECTILE; ++i)
    {
        player_projectile[i] = NULL;
    }
    
    // Reload with Save File
    player_reload(pSaveFile);
}

Player::~Player()
{
    player_delete_all_projectiles();
}

u8 Player::player_get_collision_facing()
{
    // Get Collision Blocked
    if (player_facing < FACING_LEFT)
        return COLLISION_RIGHT;
    else if (player_facing > FACING_RIGHTDOWN && player_facing < FACING_UP)
        return COLLISION_LEFT;
    else if (player_facing == FACING_UP)
        return COLLISION_UP;
    else if (player_facing == FACING_DOWN)
        return COLLISION_DOWN;
    
    return COLLISION_NONE;
}

void Player::player_save(SaveFile* pSaveFile)
{
    // Delete Player Projectiles
    player_delete_all_projectiles();
    
    // Save Player Values
    pSaveFile->savefile_set_x(player_x);
    pSaveFile->savefile_set_y(player_y);
    pSaveFile->savefile_set_barriersuit(player_barriersuit);
    pSaveFile->savefile_set_metroids(player_metroids);
    pSaveFile->savefile_set_phase(player_phase);
    pSaveFile->savefile_set_health(player_health);
    pSaveFile->savefile_set_healthtanks(player_healthtanks_total);
    pSaveFile->savefile_set_healthtanks_available(player_healthtanks_available);
    pSaveFile->savefile_set_missles_available(player_missles_available);
    pSaveFile->savefile_set_missles_total(player_missles_total);
}

void Player::player_reload(SaveFile* pSaveFile)
{
    // Animation
    player_state = PLAYERSTATE_IDLE;
    player_anim = PLAYERANIM_FRONT;
    player_x = pSaveFile->savefile_get_x();
    player_y = pSaveFile->savefile_get_y();
    player_speed = DEFAULT_PLAYER_SPEED;
    
    // Facing
    player_facing = FACING_RIGHT;
    player_shoot_facing = FACING_RIGHT;

    // Dimensions
    player_width = PLAYER_SPRITEFRAME_W / 2;
    player_height = PLAYER_SPRITEFRAME_H - (PLAYER_SPRITEFRAME_H / 10);
    player_collision = COLLISION_NONE;
    
    // Player Damage
    player_invincibility_frames = 0;
    player_damage_frames = 0;
    
    // Projectiles General
    player_charge_beam = 0;
    player_projectiles_count = 0;
    player_shooting_lock_frames = 0;
    
    player_took_damage = false;
    player_dead = false;
    player_shooting = false;
    player_morphball = false;
    player_morphball_running_jump = false;
    player_jump_movement_cancel = false;
    player_crouched = false;
    player_spiderball = false;
    player_barriersuit = pSaveFile->savefile_get_barriersuit();
    
    // Health
    player_health = pSaveFile->savefile_get_health();
    player_healthtanks_total = pSaveFile->savefile_get_healthtanks();
    player_healthtanks_available = pSaveFile->savefile_get_healthtanks_available();
    
    // Missles
    player_missles_total = pSaveFile->savefile_get_missles_total();
    player_missles_available = pSaveFile->savefile_get_missles_available();
    
    // Story
    player_metroids = pSaveFile->savefile_get_metroids();
    player_phase = pSaveFile->savefile_get_phase();
    
    // Projectiles
    for (u8 i = 0; i < PLAYER_MAX_PROJECTILE; ++i)
    {
        player_projectile[i] = NULL;
    }
}

void Player::player_set_controls(bool movement_changed, bool shooting_changed)
{
    /* Set Movement Keys */
    
    if (movement_changed == true)
    {
        // Movement Changed to DPAD
        movement_up = KEY_DUP;
        movement_down = KEY_DDOWN;
        movement_left = KEY_DLEFT;
        movement_right = KEY_DRIGHT;
        
        beam_up = KEY_CPAD_UP;
        beam_down = KEY_CPAD_DOWN;
        beam_left = KEY_CPAD_LEFT;
        beam_right = KEY_CPAD_RIGHT;
    }
    else
    {
        // Default Movement
        movement_up = KEY_CPAD_UP;
        movement_down = KEY_CPAD_DOWN;
        movement_left = KEY_CPAD_LEFT;
        movement_right = KEY_CPAD_RIGHT;
        
        beam_up = KEY_DUP;
        beam_down = KEY_DDOWN;
        beam_left = KEY_DLEFT;
        beam_right = KEY_DRIGHT;
    }
    
    /* Set Shooting Key */
    
    if (shooting_changed == true)
    {
        // Shooting Changed to Y
        shooting_button = KEY_Y;
    }
    else
    {
        // Default Shooting - B
        shooting_button = KEY_B;
    }
}

void Player::player_update_hitbox()
{
    if (player_crouched == true)
    {
        player_hitbox_height = PLAYER_HITBOX_CROUCHED_H;
        player_hitbox_width = PLAYER_HITBOX_CROUCHED_W;
        player_hitbox_x = player_x - (PLAYER_HITBOX_CROUCHED_W / 2);
    }
    else if (player_morphball == true)
    {
        player_hitbox_height = PLAYER_HITBOX_MORPHBALL_H;
        player_hitbox_width = PLAYER_HITBOX_MORPHBALL_W;
        player_hitbox_x = player_x - (PLAYER_HITBOX_MORPHBALL_W / 2);
    }
    else
    {
        player_hitbox_height = PLAYER_HITBOX_STANDING_H;
        player_hitbox_width = PLAYER_HITBOX_STANDING_W;
        player_hitbox_x = player_x - (PLAYER_HITBOX_STANDING_W / 2);
    }
}

void Player::player_checkmovement_jumping(Sound* pSound, u8 collision, bool is_running, SaveFile* pSaveFile)
{
    // Jumping
    if (collision & COLLISION_UP)
    {/*Collision Up = No Jumping*/}
    else
    {
        if (is_running == false)
        {
            if (player_morphball == false)
            {
                // Set Animation
                player_anim = PLAYERANIM_JUMP90;
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_RUN, SOUNDTYPE_PLAYER_MOVING, false, true);
                
                // Reset Crouched
                player_crouched = false;
                
                // Set State
                player_state = PLAYERSTATE_JUMPING;
            }
            else if (player_morphball == true && pSaveFile->savefile_get_springball() == true)
            {
                // Set Animation
                player_anim = PLAYERANIM_MORPHBALL;
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_MORPHBALL, SOUNDTYPE_PLAYER_MOVING, false, true);
                
                // Set State
                player_morphball_running_jump = false;
                player_state = PLAYERSTATE_JUMPING;
            }
        }
        else
        {
            if (player_morphball == false)
            {
                // Set Animation
                if (pSaveFile->savefile_get_screwattack() == false && player_anim != PLAYERANIM_JUMPSPIN)
                    player_anim = PLAYERANIM_JUMPSPIN;
                else if (pSaveFile->savefile_get_screwattack() == true && player_anim != PLAYERANIM_JUMPSCREW)
                    player_anim = PLAYERANIM_JUMPSCREW;
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_JUMP_SPIN, SOUNDTYPE_PLAYER_MOVING, false, true);
                
                // Reset Crouched
                player_crouched = false;
                
                // Set State
                player_state = PLAYERSTATE_JUMPING;
                player_morphball_running_jump = false;
            }
            else if (player_morphball == true && pSaveFile->savefile_get_springball() == true)
            {
                // Set Anim
                player_anim = PLAYERANIM_MORPHBALL;
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_MORPHBALL, SOUNDTYPE_PLAYER_MOVING, false, true);

                // Set State
                player_morphball_running_jump = true;
                player_state = PLAYERSTATE_JUMPING;
            }
        }
    }
}

void Player::player_checkmovement_running(Sound* pSound, u8 collision, bool is_in_water)
{
    // Running
    if (player_morphball == true && player_anim != PLAYERANIM_MORPHBALL)
        player_anim = PLAYERANIM_MORPHBALL;
    else if (player_morphball == false)
    {
        // Change Animation
        if (player_anim != PLAYERANIM_RUN)
            player_anim = PLAYERANIM_RUN;
    
        // Play Sound
        if (player_health > DEFAULT_LOW_HEALTH)
        {
            if (is_in_water == false)
            {
                pSound->sound_play_fx(SOUND_PLAYER_RUN, SOUNDTYPE_PLAYER_MOVING, false, false);
            }
            else
            {
                pSound->sound_play_fx(SOUND_PLAYER_WADE, SOUNDTYPE_PLAYER_MOVING, false, false);
            }
        }
    }
    
    // Set State
    player_state = PLAYERSTATE_RUNNING;
}

void Player::player_checkmovement_change(Sound* pSound, u8 collision, bool moved_up, bool spiderball, SaveFile* pSaveFile)
{
    if (moved_up == true)
    {
        // Check Current State
        if (player_morphball == true && player_crouched == false)
        {
            player_crouched = true;
            player_morphball = false;
            player_spiderball = false;
                
            // Play Sound
            pSound->sound_play_fx(SOUND_PLAYER_STANDINGMOVE, SOUNDTYPE_PLAYER_MOVING, false, true);
                
            // Set State
            player_anim = PLAYERANIM_CROUCHED;
            player_state = PLAYERSTATE_IDLE;
        }
        else if (player_crouched == true && player_morphball == false)
        {
            if (collision & COLLISION_UP)
            {}
            else
            {
                player_crouched = false;
                player_morphball = false;
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_STANDINGMOVE, SOUNDTYPE_PLAYER_MOVING, false, true);
                
                // Set State
                player_anim = PLAYERANIM_STANDING;
                player_state = PLAYERSTATE_IDLE;
            }
        }
    }
    else
    {
        // Check Current State
        if (player_crouched == false && player_morphball == false)
        {
            player_crouched = true;
            
            // Play Sound
            pSound->sound_play_fx(SOUND_PLAYER_STANDINGMOVE, SOUNDTYPE_PLAYER_MOVING, false, true);
         
            // Set State
            player_anim = PLAYERANIM_CROUCHED;
            player_state = PLAYERSTATE_IDLE;
        }
        else if (player_crouched == true && player_morphball == false)
        {
            player_crouched = false;
            player_morphball = true;
            
            // Play Sound
            pSound->sound_play_fx(SOUND_PLAYER_MORPHBALL, SOUNDTYPE_PLAYER_MOVING, false, true);
            
            // Set State
            player_anim = PLAYERANIM_MORPHBALL;
            player_state = PLAYERSTATE_IDLE;
        }
    }
}

ProjectileType Player::player_check_projectile_type(u8 projectile_used, bool projectile_charged)
{
    switch (projectile_used)
    {
        case BEAM_STANDARD:
            if (projectile_charged == true)
                return PROJECTILE_POWERBEAM_CHARGED;
            else
                return PROJECTILE_POWERBEAM;
            break;
        case BEAM_ICE:
            if (projectile_charged == true)
                return PROJECTILE_ICEBEAM_CHARGED;
            else
                return PROJECTILE_ICEBEAM;
            break;
        case BEAM_WAVE:
            if (projectile_charged == true)
                return PROJECTILE_WAVEBEAM_CHARGED;
            else
                return PROJECTILE_WAVEBEAM;
            break;
        case BEAM_SPAZER:
            if (projectile_charged == true)
                return PROJECTILE_SPAZERBEAM_CHARGED;
            else
                return PROJECTILE_SPAZERBEAM;
            break;
        case BEAM_PLASMA:
            if (projectile_charged == true)
                return PROJECTILE_PLASMABEAM_CHARGED;
            else
                return PROJECTILE_PLASMABEAM;
            break;
        case USEDPROJECTILE_MISSLE:
            return PROJECTILE_MISSLE;
            break;
        case USEDPROJECTILE_BOMB:
            return PROJECTILE_BOMB;
            break;
            
        default:
            return PROJECTILE_POWERBEAM;
            break;
    }
}

void Player::player_check_shooting_lock()
{
    // Update Shooting Lock
    if (player_shooting == true)
    {
        if (player_shooting_lock_frames <= 0)
        {
            player_last_projectile = 99;
            player_shooting = false;
        }
        else
            --player_shooting_lock_frames;
        
        return;
    }
}

void Player::player_checkshooting(Sound* pSound, bool is_missle, SaveFile* pSaveFile)
{
    /* Check for wrong Animation */
    
    if (player_anim == PLAYERANIM_CLIMBING || player_anim == PLAYERANIM_DAMAGE || player_anim == PLAYERANIM_JUMPSCREW || player_anim == PLAYERANIM_JUMPSPIN)
    {
        return;
    }

    /* Check Shooting Lock */
    
    if (player_shooting == true)
    {
        return;
    }

    // Reset Projectile
    player_last_projectile = 99;
    
    /* Spawn Projectile */
    
    if (player_morphball == true)
    {
        /* Morphball Bombs */
        
        if (pSaveFile->savefile_get_bombs())
        {
            // Spawn Bomb
            player_add_projectile(player_check_projectile_type(USEDPROJECTILE_BOMB, false));
            
            // Play Sound
            pSound->sound_play_fx(SOUND_PLAYER_MORPHBALL, SOUNDTYPE_PLAYER_SHOOTING, false, true);
            
            // Set Anim
            player_last_projectile = 99;
            
            // Set Shooting
            player_shooting = true;
            player_shooting_lock_frames = DEFAULT_PLAYER_SHOOTING_LOCK;
        }
    }
    else
    {
        if (is_missle == true)
        {
            /* Missles */
            
            if (player_missles_available > 0)
            {
                // Decrement Missles
                --player_missles_available;
                
                // Spawn Missle
                player_add_projectile(player_check_projectile_type(USEDPROJECTILE_MISSLE, false));
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_EXPLOSION, SOUNDTYPE_PLAYER_SHOOTING, false, true);
                
                // Set last fired Projectile
                player_last_projectile = PROJECTILE_MISSLE;
                
                // Set Shooting
                player_shooting = true;
                player_shooting_lock_frames = DEFAULT_PLAYER_SHOOTING_LOCK;
            }
        }
        else
        {
            // Set Last Fired Projectile
            player_last_projectile = pSaveFile->savefile_get_activebeam();
            
            if (player_charge_beam >= DEFAULT_PLAYER_CHARGE_BEAM)
            {
                // Spawn Charge Beam
                player_add_projectile(player_check_projectile_type(pSaveFile->savefile_get_activebeam(), true));
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_CHARGE_SHOOT, SOUNDTYPE_PLAYER_SHOOTING, false, true);
                
                // Set Shooting
                player_shooting = true;
                player_shooting_lock_frames = DEFAULT_PLAYER_SHOOTING_LOCK;
            }
            else
            {
                // Spawn Beam (Small)
                player_add_projectile(player_check_projectile_type(pSaveFile->savefile_get_activebeam(), false));
                
                // Play Sound
                pSound->sound_play_fx(SOUND_PLAYER_SHOOT, SOUNDTYPE_PLAYER_SHOOTING, false, true);
                
                // Set Shooting
                player_shooting = true;
                player_shooting_lock_frames = DEFAULT_PLAYER_SHOOTING_LOCK;
            }
        }
    }
    
    // Reset Charge Beam Counter
    player_charge_beam = 0;
}

void Player::player_checkfacing(u32 kdown, u32 kheld)
{
    if (player_state == PLAYERSTATE_LEDGE)
    {
        // Movement while hanging on Ledges
        if (player_facing == FACING_RIGHT || player_facing == FACING_RIGHTUP || player_facing == FACING_RIGHTDOWN)
        {
            if (kdown & movement_up)
            {
                // Aim Up
                player_shoot_facing = FACING_UP;
            }
            else if (kdown & movement_down)
            {
                // Aim Down
                player_shoot_facing = FACING_DOWN;
            }
            else if (kheld & movement_left)
            {
                // Aim Left
                player_shoot_facing = FACING_LEFT;
            }
        }
        else
        {
            if (kdown & movement_up)
            {
                // Aim Up
                player_shoot_facing = FACING_UP;
            }
            else if (kdown & movement_right)
            {
                // Aim Right
                player_shoot_facing = FACING_RIGHT;
            }
            else if (kdown & movement_down)
            {
                // Aim Down
                player_shoot_facing = FACING_DOWN;
            }
        }
    }
    else
    {
        // Change Facing
        if (kheld & movement_left)
        {
            if (kheld & movement_up)
                player_facing = FACING_LEFTUP;
            else if (kheld & movement_down)
                player_facing = FACING_LEFTDOWN;
            else
                player_facing = FACING_LEFT;
        }
        else if (kheld & movement_right)
        {
            if (kheld & movement_up)
                player_facing = FACING_RIGHTUP;
            else if (kheld & movement_down)
                player_facing = FACING_RIGHTDOWN;
            else
                player_facing = FACING_RIGHT;
        }
        else if (kheld & movement_up)
            player_facing = FACING_UP;
        
        // Set State
        player_state = PLAYERSTATE_IDLE;
    }
}

void Player::player_action(SaveFile* pSaveFile, Sound* pSound, u32 kdown, u32 kheld, u32 kup, u8 collision, bool is_in_water)
{
    // Static Variables
    static u8 player_climbing_sound = 0;

    /* Player Health Sounds */
    
    // Play Damage sound or Beep Sounds
    if (player_health <= DEFAULT_LOW_HEALTH)
    {
        // Update Health Sound Lock
        if (player_low_health_counter == DEFAULT_LOW_HEALTH_FRAMES)
        {
            pSound->sound_play_fx(SOUND_PLAYER_HEALTH_LOW, SOUNDTYPE_PLAYER_MOVING, false, true);
            player_low_health_counter = 0;
        }
        else
        {
            ++player_low_health_counter;
        }
    }
    else if (player_invincibility_frames == DEFAULT_INVINCIBILITY_FRAMES)
    {
        pSound->sound_play_fx(SOUND_PLAYER_DAMAGE, SOUNDTYPE_PLAYER_MOVING, false, true);
    }
    
    /* Update Player Hitbox */
    
    // Update Player Hitbox
    player_update_hitbox();
    
    /* Update Shooting Lock */
    
    // Update Shooting Lock
    player_check_shooting_lock();
    
    /* Update Special Frame Counters */
    
    // Update Invincibility Frames
    if (player_invincibility_frames > 0)
    {
        --player_invincibility_frames;
    }
    
    // Update Ledge Grab Lock
    if (player_ledge_grab_block_frames > 0)
    {
        --player_ledge_grab_block_frames;
        
        if (player_climbing_sound > 1)
        {
            player_climbing_sound = 0;
        }
    }
    
    /* Force Player Lock with damage or check input */
    
    if (player_damage_frames > 0)
    {
        // Set Anim
        player_anim = PLAYERANIM_DAMAGE;
        
        // Force Special State Reset
        player_spiderball = false;
        player_crouched = false;
        
        // Reset Last Projectile
        player_last_projectile = 99;
        
        // Count Damage Anim Frames
        --player_damage_frames;
        
        // Reset Damage Flag
        if (player_damage_frames == 0)
        {
            player_took_damage = false;
        }
    }
    else if (kheld & KEY_L && player_morphball == false && player_anim != PLAYERANIM_DAMAGE)
    {
        // Only Possible while Standing or Hanging!
        if (player_anim != PLAYERANIM_CLIMBING && player_anim != PLAYERANIM_DAMAGE && player_anim != PLAYERANIM_JUMPSCREW && player_anim != PLAYERANIM_JUMPSPIN && player_anim != PLAYERANIM_FALLING && player_anim != PLAYERANIM_FRONT && player_anim != PLAYERANIM_JUMP90)
        {
            /* Aiming */
        
            // Reset Player Animation
            if (player_anim != PLAYERANIM_HANGING && player_anim != PLAYERANIM_CROUCHED)
            {
                player_anim = PLAYERANIM_STANDING;
            }
        
            // Facing Check
            player_checkfacing(kdown, kheld);
            
            // Set Facing
            if (player_state != PLAYERSTATE_LEDGE)
            {
                player_shoot_facing = player_facing;
            }
        }
    }
    else
    {
        // Reset Facing
        if (player_facing < FACING_LEFT)
        {
            player_facing = FACING_RIGHT;
        }
        else if (player_facing < FACING_UP)
        {
            player_facing = FACING_LEFT;
        }
        
        // Reset Damage Animation
        if (player_anim == PLAYERANIM_DAMAGE && player_morphball == false)
        {
            player_anim = PLAYERANIM_STANDING;
            player_state = PLAYERSTATE_IDLE;
        }
        
        // Is the Player already jumping?
        if (player_state != PLAYERSTATE_JUMPING && player_state != PLAYERSTATE_FALLING && player_state != PLAYERSTATE_LEDGE && player_state != PLAYERSTATE_CLIMBING)
        {
            /* Running & Jumping */
            if (kheld & movement_left || kheld & movement_right)
            {
                if (collision & COLLISION_UP && player_crouched == true)
                {
                    // Set Facing
                    if (kheld & movement_left)
                        player_facing = FACING_LEFT;
                    else if (kheld & movement_right)
                        player_facing = FACING_RIGHT;
                }
                else
                {
                    // Set Facing
                    if (kheld & movement_left)
                    {
                        if (kheld & movement_up)
                        {
                            player_facing = FACING_LEFTUP;
                        }
                        else
                        {
                            player_facing = FACING_LEFT;
                        }
                    }
                    else if (kheld & movement_right)
                    {
                        if (kheld & movement_up)
                        {
                            player_facing = FACING_RIGHTUP;
                        }
                        else
                        {
                            player_facing = FACING_RIGHT;
                        }
                    }
        
                    // Jumping or Running
                    if (kdown & KEY_A)
                    {
                        // Jumping
                        player_checkmovement_jumping(pSound, collision, true, pSaveFile);
                    }
                    else
                    {
                        // Running
                        player_checkmovement_running(pSound, collision, is_in_water);
                    }
                
                    // Force remove chrouched Flag
                    player_crouched = false; // Reset Crouched
                }
            }
            else if (kdown & KEY_A)
            {
                player_checkmovement_jumping(pSound, collision, false, pSaveFile);
            }
            else if (kdown & movement_up || kdown & movement_down)
            {
                if (kdown & movement_down)
                {
                    player_checkmovement_change(pSound, collision, false, player_spiderball, pSaveFile);
                }
                else
                {
                    player_checkmovement_change(pSound, collision, true, player_spiderball, pSaveFile);
                }
            }
            else
            {
                // Set State
                player_state = PLAYERSTATE_IDLE;
        
                // Swap Anim
                if (player_anim == PLAYERANIM_RUN)
                    player_anim = PLAYERANIM_STANDING;
            }
            
            player_shoot_facing = player_facing;
        }
        else if (player_state == PLAYERSTATE_CLIMBING)
        {
            // Set Animation and State
            player_state = PLAYERSTATE_CLIMBING;
            player_anim = PLAYERANIM_CLIMBING;
            
            // Force Invincibility Frames
            player_invincibility_frames = 5;
            
            // Force Shooting Lock
            player_shooting_lock_frames = 5;
        }
        else if (player_state == PLAYERSTATE_LEDGE)
        {
            // Play Ledge Grab Sound
            if (player_climbing_sound == 0)
            {
                pSound->sound_play_fx(SOUND_PLAYER_LEDGE, SOUNDTYPE_PLAYER_MOVING, false, true);
                ++player_climbing_sound;
            }
            
            // Movement while hanging on Ledges
            if ((player_facing == FACING_RIGHT || player_facing == FACING_RIGHTUP || player_facing == FACING_RIGHTDOWN) && player_ledge_grab_block_frames == 0)
            {
                if (kheld & movement_right || kheld & movement_up)
                {
                    // Set Facing
                    player_facing = FACING_RIGHT;
                    player_shoot_facing = FACING_RIGHT;
                    
                    // Play Climb Up Sound
                    if (player_climbing_sound == 1)
                    {
                        pSound->sound_play_fx(SOUND_PLAYER_STANDINGMOVE, SOUNDTYPE_PLAYER_MOVING, false, true);
                        ++player_climbing_sound;
                    }
                    
                    // Climb over Ledge
                    player_state = PLAYERSTATE_CLIMBING;
                    player_anim = PLAYERANIM_CLIMBING;
                    
                    // Set Ledge Grab Block Frames
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
                else if (kheld & movement_down)
                {
                    // Fall Down
                    player_state = PLAYERSTATE_IDLE;
                    
                    // Set Ledge Grab Block Frames
                    ++player_climbing_sound;
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
                else if (kdown & KEY_A)
                {
                    // Jumping from Ledge
                    if (kheld & movement_left)
                    {
                        // Set Facing
                        player_facing = FACING_LEFT;
                        
                        // Jumping
                        player_checkmovement_jumping(pSound, collision, true, pSaveFile);
                    }
                    else
                    {
                        // Jump Up
                        player_checkmovement_jumping(pSound, collision, false, pSaveFile);
                    }
                    
                    // Set Ledge Grab Block Frames
                    ++player_climbing_sound;
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
            }
            else if (player_ledge_grab_block_frames == 0)
            {
                if (kheld & movement_left || kheld & movement_up)
                {
                    // Set Facing
                    player_facing = FACING_LEFT;
                    player_shoot_facing = FACING_LEFT;
                    
                    // Play Climb Up Sound
                    if (player_climbing_sound == 1)
                    {
                        pSound->sound_play_fx(SOUND_PLAYER_STANDINGMOVE, SOUNDTYPE_PLAYER_MOVING, false, true);
                        ++player_climbing_sound;
                    }
                    
                    // Climb over Ledge
                    player_state = PLAYERSTATE_CLIMBING;
                    player_anim = PLAYERANIM_CLIMBING;
                    
                    // Set Ledge Grab Block Frames
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
                else if (kheld & movement_down)
                {
                    // Fall Down
                    player_state = PLAYERSTATE_IDLE;
                    
                    // Set Ledge Grab Block Frames
                    ++player_climbing_sound;
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
                else if (kdown & KEY_A)
                {
                    // Jumping from Ledge
                    if (kheld & movement_right)
                    {
                        // Set Facing
                        player_facing = FACING_RIGHT;
                        
                        // Jumping
                        player_checkmovement_jumping(pSound, collision, true, pSaveFile);
                    }
                    else
                    {
                        // Jump Up
                        player_checkmovement_jumping(pSound, collision, false, pSaveFile);
                    }
                    
                    // Set Ledge Grab Block Frames
                    ++player_climbing_sound;
                    player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
                }
            }
        }
        else if (player_anim == PLAYERANIM_JUMPSCREW || player_anim == PLAYERANIM_JUMPSPIN)
        {
            // Fall once we release the A key
            if (kup & KEY_A)
            {
                // Set Falling State
                player_state = PLAYERSTATE_FALLING;
            }
            
            // Change facing and fall if we changed our jumping direction
            if (kdown & movement_left && player_facing != FACING_LEFT)
            {
                // Change Facing
                player_facing = FACING_LEFT;
                
                // Set Falling State
                player_state = PLAYERSTATE_FALLING;
            }
            else if (kdown & movement_right && player_facing != FACING_RIGHT)
            {
                // Change Facing
                player_facing = FACING_RIGHT;
                
                // Set Falling State
                player_state = PLAYERSTATE_FALLING;
            }
        }
        else if (player_anim == PLAYERANIM_FALLING || player_anim == PLAYERANIM_JUMP90)
        {
            // Remove Crouching Flag
            if (player_crouched == true)
            {
                player_crouched = false;
            }
            
            // Fall once we release the A key
            if (kup & KEY_A)
            {
                // Set Falling State
                player_state = PLAYERSTATE_FALLING;
            }
            
            // Facing and Movement
            if (kheld & movement_left)
            {
                // Set Facing
                player_facing = FACING_LEFT;
                player_shoot_facing = FACING_LEFT;
                    
                // Set Movement Flag
                player_jump_movement_cancel = true;
                    
                if (player_anim == PLAYERANIM_JUMP90 && player_state == PLAYERSTATE_JUMPING)
                {
                    player_state = PLAYERSTATE_FALLING;
                }
            }
            else if (kheld & movement_right)
            {
                // Set Facing
                player_facing = FACING_RIGHT;
                player_shoot_facing = FACING_RIGHT;
                    
                // Set Movement Flag
                player_jump_movement_cancel = true;
                
                
                if (player_anim == PLAYERANIM_JUMP90 && player_state == PLAYERSTATE_JUMPING)
                {
                    player_state = PLAYERSTATE_FALLING;
                }
            }
            else if (kheld & movement_up)
            {
                if (player_shoot_facing == FACING_RIGHT || player_shoot_facing == FACING_RIGHTDOWN || player_shoot_facing == FACING_RIGHTUP)
                {
                    player_shoot_facing = FACING_RIGHTUP;
                }
                else if (player_shoot_facing == FACING_LEFT || player_shoot_facing == FACING_LEFTDOWN || player_shoot_facing == FACING_LEFTUP)
                {
                    player_shoot_facing = FACING_LEFTUP;
                }
            }
            else if (kheld & movement_down)
            {
                if (player_shoot_facing == FACING_RIGHT || player_shoot_facing == FACING_RIGHTDOWN || player_shoot_facing == FACING_RIGHTUP)
                {
                    player_shoot_facing = FACING_RIGHTDOWN;
                }
                else if (player_shoot_facing == FACING_LEFT || player_shoot_facing == FACING_LEFTDOWN || player_shoot_facing == FACING_LEFTUP)
                {
                    player_shoot_facing = FACING_LEFTDOWN;
                }
            }
        }
    }
    
    /* Beam Switching */
    
    if (kdown & beam_down)
    {
        if (pSaveFile->savefile_get_spazerbeam() == true)
            pSaveFile->savefile_set_activebeam(BEAM_SPAZER);
        
        player_charge_beam = 0;
    }
    else if (kdown & beam_up)
    {
        if (pSaveFile->savefile_get_icebeam() == true)
            pSaveFile->savefile_set_activebeam(BEAM_ICE);
        
        player_charge_beam = 0;
    }
    else if (kdown & beam_left)
    {
        if (pSaveFile->savefile_get_wavebeam() == true)
            pSaveFile->savefile_set_activebeam(BEAM_WAVE);
        
        player_charge_beam = 0;
    }
    else if (kdown & beam_right)
    {
        if (pSaveFile->savefile_get_plasmabeam() == true)
            pSaveFile->savefile_set_activebeam(BEAM_PLASMA);
        
        player_charge_beam = 0;
    }
    
    /* Shooting */
    
    if (player_damage_frames == 0)
    {
        if (kdown & KEY_R)
        {
            if (player_anim != PLAYERANIM_CLIMBING && player_anim != PLAYERANIM_DAMAGE && player_anim != PLAYERANIM_JUMPSCREW && player_anim != PLAYERANIM_JUMPSPIN && player_state != PLAYERSTATE_CLIMBING && player_missles_available > 0)
            {
                player_checkshooting(pSound, true, pSaveFile);
                player_charge_beam = 0;
                player_shooting = true;
            }
        }
        else if (kheld & shooting_button && player_charge_beam <= DEFAULT_PLAYER_CHARGE_BEAM)
        {
            // Morphball
            if (player_morphball == false && player_state != PLAYERSTATE_CLIMBING)
            {
                // Play Sound
                if (player_charge_beam == 0)
                {
                    pSound->sound_play_fx(SOUND_PLAYER_CHARGE_START, SOUNDTYPE_PLAYER_SHOOTING, false, true);
                }
                else
                {
                    pSound->sound_play_fx(SOUND_PLAYER_CHARGE_ONGOING, SOUNDTYPE_PLAYER_SHOOTING, false, false);
                }
        
                // Set Charge Beam
                if (player_charge_beam < DEFAULT_PLAYER_CHARGE_BEAM)
                {
                    ++player_charge_beam;
                }
        
                if (player_anim != PLAYERANIM_CLIMBING && player_anim != PLAYERANIM_DAMAGE && player_anim != PLAYERANIM_JUMPSCREW && player_anim != PLAYERANIM_JUMPSPIN)
                {
                    player_shooting = true;
                }
                
                // Set Beam
                player_last_projectile = pSaveFile->savefile_get_activebeam();
            }
            else
            {
                player_last_projectile = pSaveFile->savefile_get_activebeam();
                player_charge_beam = 0;
            }
        }
        else if (kup & shooting_button)
        {
            if (player_anim != PLAYERANIM_CLIMBING && player_anim != PLAYERANIM_DAMAGE && player_anim != PLAYERANIM_JUMPSCREW && player_anim != PLAYERANIM_JUMPSPIN && player_state != PLAYERSTATE_CLIMBING)
            {
                // Check for Beams
                player_checkshooting(pSound, false, pSaveFile);
            }
            else
            {
                // We played Sound
                if (player_charge_beam > 0)
                {
                    // Stop Sound
                    pSound->sound_stop(SOUNDTYPE_PLAYER_SHOOTING);
                }
                
                // Reset Charge Beam Counter
                player_charge_beam = 0;
            }
        }
        else
        {
            // Reset Charge Beam Counter
            player_charge_beam = 0;
        }
    }

    /* Spider Ball */
    
    if (player_damage_frames == 0)
    {
        if (kheld & KEY_X && pSaveFile->savefile_get_spiderball() == true && player_morphball == true)
        {
            player_spiderball = true;
        }
        else if (kup & KEY_X && pSaveFile->savefile_get_spiderball() == true)
        {
            player_spiderball = false;
        }
    }
}

void Player::player_damage(u8 damage)
{
    // Don't perform for 0 Damage!
    if (damage == 0)
        return;
    
    // Variable for Health Calculation
    s32 health_result;
    
    // Damage only if no invincibility Frames are up
    if (player_invincibility_frames == 0)
    {
        // Charge Beam Reset
        player_charge_beam = 0;
        
        // Get Full Player Health
        health_result = (player_health + (player_healthtanks_available * DEFAULT_HEALTHTANK_SIZE)) - damage;
        
        // Calculation Result Actions
        if (health_result <= 0)
        {
            // No Health left - Player died
            player_healthtanks_available = 0;
            player_health = 0;
            player_dead = true;
        }
        else
        {
            // Set Health Tanks
            while (damage >= DEFAULT_HEALTHTANK_SIZE)
            {
                --player_healthtanks_available;
                damage -= DEFAULT_HEALTHTANK_SIZE;
            }
            
            // Set Health
            if (player_health - damage < 0)
            {
                // Set Damage
                damage -= player_health;
                
                // Set Health Tanks
                --player_healthtanks_available;
                
                // Set Health
                player_health = DEFAULT_HEALTHTANK_SIZE - damage;
                
            }
            else
            {
                // Calculate Normally
                player_health -= damage;
            }
        }
        
        // Set Invincibility Frames
        player_invincibility_frames = DEFAULT_INVINCIBILITY_FRAMES;
        
        // Set Anim
        if (player_anim != PLAYERANIM_JUMP90 && player_anim != PLAYERANIM_JUMPSPIN && player_anim != PLAYERANIM_JUMPSCREW)
        {
            player_anim = PLAYERANIM_DAMAGE;
        }
        
        // Fix State
        if (player_state == PLAYERSTATE_RUNNING || player_state == PLAYERSTATE_LEDGE)
        {
            // Set State
            player_state = PLAYERSTATE_IDLE;
        }
        else if (player_state == PLAYERSTATE_JUMPING || player_state == PLAYERSTATE_FALLING)
        {
            // Set State
            player_state = PLAYERSTATE_FALLING;
            
            // Set Anim
            player_anim = PLAYERANIM_FALLING;
        }
        
        // Set Ledge Grab Lock Frames
        player_ledge_grab_block_frames = DEFAULT_LEDGE_BLOCK_FRAMES;
        
        // Set Damage Frames
        player_damage_frames = DEFAULT_DAMAGE_FRAMES;
        
        // Set Damage Flag
        player_took_damage = true;
    }
}

void Player::player_heal(u8 heal)
{
    // New Health
    s32 health_result = (player_health + (player_healthtanks_available * DEFAULT_HEALTHTANK_SIZE)) + heal;
    
    // Set Health and Healthtanks
    if (health_result >= player_health + (player_healthtanks_total * DEFAULT_HEALTHTANK_SIZE))
    {
        // Set Health to full
        player_health = DEFAULT_FULL_HEALTH;
        player_healthtanks_available = player_healthtanks_total;
    }
    else
    {
        // Set Health Tanks
        while (health_result >= DEFAULT_HEALTHTANK_SIZE)
        {
            ++player_healthtanks_available;
            health_result -= DEFAULT_HEALTHTANK_SIZE;
        }
        
        // Set Health
        if (player_health + heal > DEFAULT_FULL_HEALTH)
        {
            // Set Damage
            heal -= DEFAULT_FULL_HEALTH - player_health;
            
            // Set Health Tanks
            if (player_healthtanks_available < player_healthtanks_total)
            {
                ++player_healthtanks_available;
            }
            
            // Set Health
            player_health = heal - 1;
            
        }
        else
        {
            // Calculate Normally
            player_health += heal;
        }
    }
}

void Player::player_increase_available_missles(u8 missles)
{
    if (player_missles_available + missles > player_missles_total)
    {
        player_missles_available = player_missles_total;
    }
    else
    {
        player_missles_available += missles;
    }
}

void Player::player_increase_total_missles(u8 missles)
{
    if (player_missles_total + missles > DEFAULT_MAX_MISSLES_TOTAL)
    {
        player_missles_total = DEFAULT_MAX_MISSLES_TOTAL;
    }
    else
    {
        player_missles_total += missles;
    }
}

void Player::player_heal_full()
{
    // Reset Health
    player_health = DEFAULT_FULL_HEALTH;
    player_healthtanks_available = player_healthtanks_total;
}

void Player::player_increase_healthtanks()
{
    // Increase Health Tanks
    ++player_healthtanks_total;
    
    // Heal Health
    player_heal_full();
}

bool Player::player_collision_in_range(s32 actor_a_value, s32 actor_b_min, s32 actor_b_max)
{
    // Return Compare Result
    return (actor_a_value >= actor_b_min) && (actor_a_value <= actor_b_max);
}

bool Player::player_get_enemy_collision(s32 enemy_x, s32 enemy_y, s32 enemy_w, s32 enemy_h)
{
    // Compare Positions
    bool x_overlap = player_collision_in_range(player_x, enemy_x, enemy_x + enemy_w) ||
                     player_collision_in_range(enemy_x, player_x, player_x - (player_hitbox_width / 2) + player_hitbox_width);
    
    bool y_overlap = player_collision_in_range(player_y, enemy_y, enemy_y + enemy_h) ||
                     player_collision_in_range(enemy_y, player_y, player_y + player_hitbox_height);
    
    // Return Result
    if (x_overlap == true && y_overlap == true)
    {
        return true;
    }
    else
    {
        return false;
    }
}

void Player::player_add_projectile(ProjectileType projectile_type)
{
    // Projectile Start Coordinates
    s32 x = 0, y = 0;

    /* Set Spawn Position */
    
    x = player_x;
    
    // Running and Aiming Up Corrections
    if (player_anim == PLAYERANIM_RUN)
    {
        if (player_facing == FACING_LEFTUP)
        {
            x -= PROJECTILE_AIM_UP_CORRECTION;
        }
        else if (player_facing == FACING_RIGHTUP)
        {
            x += PROJECTILE_AIM_UP_CORRECTION;
        }
    }
    
    if (player_facing == FACING_LEFTDOWN || player_facing == FACING_RIGHTDOWN)
    {
        if (player_crouched == true)
            y = player_y + ((player_hitbox_height / 2) - PROJECTILE_HEIGHT_HALF_CROUCHED);
        else
            y = player_y + ((player_hitbox_height / 2) - PROJECTILE_HEIGHT_HALF_STANDING);
    }
    if (player_facing == FACING_LEFTUP || player_facing == FACING_RIGHTUP)
    {
        if (player_crouched == true)
            y = player_y + (player_hitbox_height - PROJECTILE_HEIGHT_HALF_CROUCHED);
        else
            y = player_y + (player_hitbox_height - PROJECTILE_HEIGHT_HALF_STANDING);
    }
    else if (player_facing == FACING_UP)
    {
        y = player_y + player_hitbox_height;
    }
    else if (player_facing == FACING_DOWN)
    {
        y = player_y;
    }
    else
    {
        if (player_crouched == true)
            y = player_y + (player_hitbox_height - PROJECTILE_HEIGHT_HALF_CROUCHED);
        else
            y = player_y + (player_hitbox_height - PROJECTILE_HEIGHT_HALF_STANDING);
    }
    
    /* Set Launch State */
    
    ProjectileLaunchState launch_state = PROJECTILE_LAUNCHSTATE_STANDING;
    
    if (player_crouched == true)
    {
        launch_state = PROJECTILE_LAUNCHSTATE_CROUCHED;
    }
    else if (player_morphball == true)
    {
        launch_state = PROJECTILE_LAUNCHSTATE_MORPHBALL;
    }
    
    /* Correct Facing */
    
    u8 facing = player_shoot_facing;
    
    if (player_state == PLAYERSTATE_JUMPING || player_state == PLAYERSTATE_FALLING)
    {
        if (player_shoot_facing == FACING_LEFTUP || player_shoot_facing == FACING_RIGHTUP)
            facing = FACING_UP;
        else if (player_shoot_facing == FACING_LEFTDOWN || player_shoot_facing == FACING_RIGHTDOWN)
            facing = FACING_DOWN;
    }
    
    /* Create Projectile */
    
    for (u8 i = 0; i <= player_projectiles_count + 1; ++i)
    {
        if (player_projectile[i] == NULL)
        {
            // Create Object
            player_projectile[i] = new Projectile(projectile_type, launch_state, x, y, facing);
        
            // Increment if Possible
            if (i == player_projectiles_count + 1 && player_projectiles_count < PLAYER_MAX_PROJECTILE)
            {
                player_projectiles_count++;
            }
            
            // Exit Loop
            break;
        }
    }
}

Projectile* Player::player_get_projectile(u8 projectile)
{
    // Non Existent Projectiles
    if (projectile > player_projectiles_count)
        return NULL;
    
    // Return Projectile
    return player_projectile[projectile];
}

void Player::player_delete_projectile(u8 projectile)
{
    // Non Existent Projectiles
    if (projectile > player_projectiles_count)
        return;
    
    // Delete Projectile
    if (player_projectile[projectile] != NULL)
    {
        delete player_projectile[projectile];
    }
    
    // Reset Projectile Pointer
    player_projectile[projectile] = NULL;
}

void Player::player_delete_all_projectiles()
{
    // Delete Projectiles
    for (u8 i = 0; i < PLAYER_MAX_PROJECTILE; ++i)
    {
        if (player_projectile[i] != NULL)
        {
            delete player_projectile[i];
        }
        
        player_projectile[i] = NULL;
    }
    
    // Reset Projectile Count
    player_projectiles_count = 0;
}
