//
//  sprite_handler.cpp
//  3ds_metroid
//
//  Created by jbr373 on 17/02/16.
//

#include <string.h>
#include <sfil.h>

#include "sprite_handler.h"
#include "actor_facing.h"
#include "filepaths.h"

#define DEFAULT_FRAME_SPEED             2

#define FRAME_MIN_RIGHT                 0
#define FRAME_MAX_RIGHT                 3

#define FRAME_MIN_LEFT                  4
#define FRAME_MAX_LEFT                  7

#define MAX_FRAME                       7

#define EFFECT_SPRITE_W                 100
#define EFFECT_SPRITE_H                 90
#define EFFECT_SPRITESHEET_W            500
#define EFFECT_SPRITESHEET_H            450

enum EffectsSpriteLines
{
    EFFECT_JUMP_SHADOW = 0,
};

SpriteHandler::SpriteHandler(SaveFile* pSaveFile)
{
    // Set Object Sprites NULL
    for (u8 i = 0; i < MAX_SPRITE_SLOTS; ++i)
    {
        pActorSprite[i] = NULL;
    }
    
    // Load Effects
    pTexture_Effects = NULL;
    spritehandler_load_effects();
    
    playersprite_load_sprites(false);
}

SpriteHandler::~SpriteHandler()
{
    playersprite_clear_sprites();
}

/***********************************/
/*         SPRITE RETURN           */
/***********************************/

/* Effect Sprites */

SpriteSheet* SpriteHandler::spritehandler_get_effect_sprite(Player* pPlayer)
{
    // Set Effect Sprite by Animation!
    if (pPlayer->player_get_anim() == PLAYERANIM_JUMP90 && pPlayer->player_get_state() == PLAYERSTATE_JUMPING)
    {
        switch (pPlayer->player_get_facing())
        {
            case FACING_LEFT:
                pTexture_Effects->spritesheet_startframe = 1;
                pTexture_Effects->spritesheet_endframe = 1;
                pTexture_Effects->spritesheet_frame = 1;
                break;
            case FACING_RIGHT:
                pTexture_Effects->spritesheet_startframe = 0;
                pTexture_Effects->spritesheet_endframe = 0;
                pTexture_Effects->spritesheet_frame = 0;
                break;
            case FACING_LEFTUP:
                pTexture_Effects->spritesheet_startframe = 3;
                pTexture_Effects->spritesheet_endframe = 3;
                pTexture_Effects->spritesheet_frame = 3;
                break;
            case FACING_RIGHTUP:
                pTexture_Effects->spritesheet_startframe = 2;
                pTexture_Effects->spritesheet_endframe = 2;
                pTexture_Effects->spritesheet_frame = 2;
                break;
                
            default:
                pTexture_Effects->spritesheet_startframe = 4;
                pTexture_Effects->spritesheet_endframe = 4;
                pTexture_Effects->spritesheet_frame = 4;
                break;
        }
        
        // Return Sprite
        return pTexture_Effects;
    }
    
    // Return No Sprite needed
    return NULL;
}

/* Player Sprites */

SpriteSheet* SpriteHandler::spritehandler_get_sprite(Player* pPlayer, bool is_in_water)
{
    // Special Handling
    if (pPlayer->player_get_anim() == PLAYERANIM_FALLING ||
        pPlayer->player_get_anim() == PLAYERANIM_JUMP90  ||
        pPlayer->player_get_anim() == PLAYERANIM_HANGING ||
        pPlayer->player_get_anim() == PLAYERANIM_CLIMBING)
    {
        return playersprite_get_sprite(pPlayer->player_get_anim(),
                                       pPlayer->player_get_shooting_facing(),
                                       pPlayer->player_get_state(),
                                       pPlayer->player_get_shooting(),
                                       is_in_water);
    }
    else
    {
        return playersprite_get_sprite(pPlayer->player_get_anim(),
                                       pPlayer->player_get_facing(),
                                       pPlayer->player_get_state(),
                                       pPlayer->player_get_shooting(),
                                       is_in_water);
    }
}

/* Actor Sprite */

SpriteSheet* SpriteHandler::spritehandler_get_sprite(u8 sprite_type, std::string sprite_path, u8 sprite_anim, u8 sprite_facing, u8 sprite_frame)
{
    // Check if we're having NONE as Sprite Path
    if (strcmp(sprite_path.c_str(), LINE_DO_NOT_USE) == 0)
    {
        return NULL;
    }
    
    // Search for Object Path
    u8 i = 0;
    while (i < MAX_SPRITE_SLOTS)
    {
        // Check if the Object is initialized
        if (pActorSprite[i] != NULL)
        {
            // Check Path String
            if (pActorSprite[i]->actorsprite_get_filepath() == sprite_path)
            {
                break;
            }
            else
                ++i;
        }
        else
            ++i;
    }
    
    // No Object Initialized
    if (i > MAX_SPRITE_SLOTS - 1)
        return NULL;

    /* Set and Load Sprite depending on State */
    
    if (sprite_type == ACTORSPRITE_TYPE_OBJECT)
    {
        // Return Standard Sprite
        return pActorSprite[i]->actorsprite_get_sprite(sprite_anim, 0, MAX_FRAME, sprite_frame);
    }
    else if (sprite_type == ACTORSPRITE_TYPE_ENEMY)
    {
        if (sprite_anim == ENEMYANIM_STANDING)
        {
            return pActorSprite[i]->actorsprite_get_sprite(0, 0, 0, 0);
        }
        else if (sprite_anim == ENEMYANIM_STANDING_DAMAGE)
        {
            return pActorSprite[i]->actorsprite_get_sprite(1, 0, 0, 0);
        }
        else
        {
            if (sprite_facing < FACING_LEFT)
            {
                return pActorSprite[i]->actorsprite_get_sprite(sprite_anim, FRAME_MIN_RIGHT, FRAME_MAX_RIGHT, sprite_frame);
            }
            else
            {
                return pActorSprite[i]->actorsprite_get_sprite(sprite_anim, FRAME_MIN_LEFT, FRAME_MAX_LEFT, sprite_frame);
            }
        }
    }
    else
        return NULL;
    
    // Not Really possible, but return NULL
    return NULL;
}

/************************************/
/*         SPRITE LOADING           */
/************************************/

void SpriteHandler::spritehandler_load_effects()
{
    // Allocate Space
    pTexture_Effects = new SpriteSheet();
    
    // Load Texture
    pTexture_Effects->spritesheet_texture = sfil_load_PNG_file(IMG_EFFECTS, SF2D_PLACE_RAM);
    
    // Set Defaults
    pTexture_Effects->spritesheet_sprite_line = EFFECT_JUMP_SHADOW;
    pTexture_Effects->spritesheet_startframe = 0;
    pTexture_Effects->spritesheet_endframe = 0;
    pTexture_Effects->spritesheet_frame = 0;
    pTexture_Effects->spritesheet_h_single = EFFECT_SPRITE_H;
    pTexture_Effects->spritesheet_w_single = EFFECT_SPRITE_W;
    pTexture_Effects->spritesheet_frame_speed_current = 0;
    pTexture_Effects->spritesheet_frame_speed = 0;
    pTexture_Effects->spritesheet_w_full = EFFECT_SPRITESHEET_W;
    pTexture_Effects->spritesheet_h_full = EFFECT_SPRITESHEET_H;
}

void SpriteHandler::spritehandler_load_sprite(std::string sprite_path, u8 sprite_max_line, s32 sprite_w, s32 sprite_h, u8 sprite_frame_speed)
{
    // Already Loaded
    bool already_loaded = false;
    
    // Check if we're having NONE as Sprite Path
    if (strcmp(sprite_path.c_str(), LINE_DO_NOT_USE) == 0)
    {
        return;
    }
    
    // Check Loaded Sprites for the Same Filepath
    u8 i = 0;
    while (i < MAX_SPRITE_SLOTS)
    {
        // Check if the Object is initialized
        if (pActorSprite[i] != NULL)
        {
            // Check Path String
            if (pActorSprite[i]->actorsprite_get_filepath() == sprite_path)
            {
                already_loaded = true;
                break;
            }
            else
                ++i;
        }
        else
            ++i;
    }
    
    // Return if already loaded
    if (already_loaded == true)
    {
        return;
    }
    
    // Reset Counter
    i = 0;
    
    // Check for Loaded Sprites
    while (i < MAX_SPRITE_SLOTS)
    {
        // Check Path String
        if (pActorSprite[i] == NULL)
        {
            // Create Class
            pActorSprite[i] = new ActorSprite();
            
            // Load Sprite
            pActorSprite[i]->actorsprite_load(sprite_path,
                                               sprite_max_line,
                                               sprite_w,
                                               sprite_h,
                                               sprite_frame_speed);
            
            // Exit Loop
            break;
        }
        else
            ++i;
    }
}

/***********************************/
/*         CLEAR SPRITES           */
/***********************************/

void SpriteHandler::spritehandler_clear_all_sprites()
{
    // Check for Loaded Sprites
    u8 i = 0;
    while (i < MAX_SPRITE_SLOTS)
    {
        // Delete Used Sprite Slot
        if (pActorSprite[i] != NULL)
        {
            if (strcmp(pActorSprite[i]->actorsprite_get_filepath().c_str(), LINE_DO_NOT_USE) != 0)
            {
                // Delete Spritesheet
                pActorSprite[i]->actorsprite_delete();
                
                // Delete Class
                delete pActorSprite[i];
                pActorSprite[i] = NULL;
            }
        }
        
        ++i;
    }
}
