//
//  stage.cpp
//  3ds_metroid
//
//  Created by jbr373 on 16/01/16.
//

#include <sstream>
#include <fstream>
#include <sfil.h>
#include <sf2d.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "stage.h"
#include "filepaths.h"
#include "stage_collision.h"
#include "actor_facing.h"

#define ROOM_ENEMY_FILE             "_enemy"
#define ROOM_ITEM_FILE              "_item"
#define ROOM_OBJECT_FILE            "_object"

#define STAGE_MAP_OBJECT_LINES      2 // 0 - 2, ID, X, Y
#define STAGE_RUN_COLLISION_H       6

#define OBJECT_MAX_FACING_RIGHT     2

#define STAGE_GET_PORT_CORRECTION   5
#define STAGE_LEFT_COL_CORRECTION   6

#define PAUSE_TEXT_ID               4

/*
 *      TODO
 * ---------------
 * - Y Up Collision Check is wierd and overcompicated (or simply wrong) - X Left/Right and Y Down seems fine
 */

struct EnemyMapData
{
    u8 stage_enemy_file_path_length[STAGE_MAX_ENEMIES];
    std::string stage_enemy_file_path[STAGE_MAX_ENEMIES];
    s32 stage_enemy_x[STAGE_MAX_ENEMIES];
    s32 stage_enemy_y[STAGE_MAX_ENEMIES];
};

struct ObjectMapData
{
    u8 stage_object_file_path_length[STAGE_MAX_OBJECTS];
    std::string stage_object_file_path[STAGE_MAX_OBJECTS];
    s32 stage_object_x[STAGE_MAX_OBJECTS];
    s32 stage_object_y[STAGE_MAX_OBJECTS];
};

struct ItemMapData
{
    u8 stage_item_file_path_length[STAGE_MAX_ITEMS];
    std::string stage_item_file_path[STAGE_MAX_ITEMS];
    s32 stage_item_x[STAGE_MAX_ITEMS];
    s32 stage_item_y[STAGE_MAX_ITEMS];
};

Stage::Stage()
{
    // Stage ID's
    stage_id = 1;
    stage_phase = 0;
    stage_minimap_id = 0;

    // File Paths
    stage_sound_path = "NONE";
    stage_tileset_path = "NONE";
    stage_background_path = "NONE";
    
    // Stage Dimensions
    stage_width = 0;
    stage_height = 0;
    stage_layout_tiles_width = 0;
    stage_layout_tiles_height = 0;
   
    // Stage Actors
    stage_enemies = 0;
    stage_items = 0;
    stage_objects = 0;
    
    for (u8 i = 0; i < STAGE_MAX_ENEMIES; ++i)
    {
        pEnemy[i] = NULL;
    }
    
    for (u8 j = 0; j < STAGE_MAX_ITEMS; ++j)
    {
        pItem[j] = NULL;
    }
    
    for (u8 k = 0; k < STAGE_MAX_OBJECTS; ++k)
    {
        pObject[k] = NULL;
    }
    
    // Image Pointers
    stage_background = NULL;
    stage_tileset = NULL;
    
    // Tileset Allignment
    stage_layout_tiles_foreground = NULL;
    stage_layout_tiles_background = NULL;
    stage_draw_tiles_background = false;
    
    // Collision
    stage_collision = NULL;
    
    // Light
    stage_light_ambient = 0;
    
    // Water
    stage_water_type = 0;
    stage_water_x = 0;
    stage_water_y = 0;
    stage_water_w = 0;
    stage_water_h = 0;
    stage_water_alpha = 0;
    stage_water_ripples_size = 0;
    stage_water_godray = false;
    stage_godray_x = 0;
    stage_godray_y = 0;
    
    // Stage Loading Flag
    stage_room_loaded = NULL;
    
    // Animation & Load State
    stage_room_change = STAGE_ROOMCHANGE_NOCHANGE;
}

Stage::~Stage()
{
    // Delete Full Stage on Class Deconstruction
    stage_delete();
}

enum LiquidTypes
{
    LIQUIDTYPE_WATER = 0,
    LIQUIDTYPE_LAVA = 1,
    LIQUIDTYPE_ACID = 2
};

/*************************************************************************************************************************/
/*                                                   Stage Loading                                                       */
/*************************************************************************************************************************/

void Stage::stage_load(Sound* pSound, SpriteHandler* pSpriteHandler, u16 room_id)
{
    /* Load a new Room */

    // Define File Path
    std::stringstream filepath;
    filepath << FILEPATH_ROOM << room_id << EXTENSIONS_MAP;
    std::string fullpath = filepath.str();
   
    // Open Map File
    std::ifstream mapfile;
    mapfile.open(fullpath, std::ios::in | std::ios::binary);
    
    // Load Room
    if (mapfile.is_open())
    {
        /**********************/
        /* Reset File Pointer */
        /**********************/
        
        // Set File Pointer to 0 (Beginning of the File)
        fileop_seekg_resetpos();
        
        /*********************************/
        /* Check Stage Magic and Version */
        /*********************************/
        
        // Check Stage File Magic
        if (fileop_read_u32(mapfile) != MAGIC_DECIMAL_MAP)
        {
            // Close File
            mapfile.close();
            
            // Kill Game
            exit(1);
        }
        
        // Set Position
        fileop_seekg_file(mapfile, sizeof(u32));
        
        // Check Stage File Version
        if (fileop_read_u32(mapfile) != VERSION_DECIMAL_MAP)
        {
            // Close File
            mapfile.close();
            
            // Kill Game
            exit(1);
        }
        
        // Set Position
        fileop_seekg_file(mapfile, sizeof(u32));
        
        /*******************************/
        /* Reset Stage Actors and Info */
        /*******************************/
        
        // Delete old Room Actors and Effects
        if (stage_room_loaded == true)
        {
            // Delete loaded Actors
            stage_delete_actors();
            
            // Reset Stage Data
            stage_reset_variables();
        }
        
        /***************************/
        /* Load Stage General Info */
        /***************************/
        
        // Stage IDs
        stage_id = fileop_read_u16(mapfile);
        fileop_seekg_file(mapfile, sizeof(u16));
        stage_phase = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Stage Dimensions
        stage_width = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        stage_layout_tiles_width = stage_width / TILESSET_TILE_DIMENSION;

        stage_height = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        stage_layout_tiles_height = stage_height / TILESSET_TILE_DIMENSION;
        
        // Actor Info
        // Get Stage Enemies
        stage_enemies = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Correct Stage Enemies if needed
        if (stage_enemies > STAGE_MAX_ENEMIES)
        {
            stage_enemies = STAGE_MAX_ENEMIES;
        }
        
        // Get Stage Objects
        stage_objects = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Correct Stage Objects if needed
        if (stage_objects > STAGE_MAX_OBJECTS)
        {
            stage_objects = STAGE_MAX_OBJECTS;
        }
        
        // Get Stage Items
        stage_items = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Correct Stage Enemies if needed
        if (stage_items > STAGE_MAX_ITEMS)
        {
            stage_items = STAGE_MAX_ITEMS;
        }
        
        // Minimap
        stage_minimap_id = fileop_read_u32(mapfile);
        fileop_seekg_file(mapfile, sizeof(u32));
        
        // Water Type for Damage / Movement Manipulation
        stage_water_type = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        /**********************/
        /* Load Stage Tileset */
        /**********************/
        
        // Get Background
        std::string background_path;
        
        // Get Background File Path Length
        u8 stage_background_path_length = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Write each char of length to string
        for (u8 i = 0; i <= stage_background_path_length; ++i)
        {
            // Get Char and set Pointer
            char c = (char)fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(stage_background_path_length));
            
            // Add char to string
            background_path += c;
        }
        
        // Get Tileset
        std::string tileset_path;
        
        // Get Tileset Path Length
        u8 stage_tileset_path_length = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Write each char of length to string
        for (u8 i = 0; i <= stage_tileset_path_length; ++i)
        {
            // Get Char and set Pointer
            char c = (char)fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(stage_tileset_path_length));
            
            // Add char to string
            tileset_path += c;
        }
        
        /*******************************/
        /* Delete old Environment Data */
        /*******************************/
        
        // Delete old Room
        if (stage_room_loaded == true)
        {
            stage_delete_environment(background_path, tileset_path);
        }
        
        /***********************/
        /* Load new Image Data */
        /***********************/
        
        // Load Stage Background
        if (strcmp(stage_background_path.c_str(), background_path.c_str()) != 0 && stage_background == NULL)
        {
            // Load Stage Background
            stage_background = sfil_load_PNG_file(background_path.c_str(), SF2D_PLACE_RAM);
            
            // Set New Sound Path
            stage_background_path = background_path;
        }
        
        // Load Stage Tileset
        if (strcmp(stage_tileset_path.c_str(), tileset_path.c_str()) != 0 && stage_tileset == NULL)
        {
            // Load Tileset
            stage_tileset = sfil_load_PNG_file(tileset_path.c_str(), SF2D_PLACE_RAM);
            
            // Set New Sound Path
            stage_tileset_path = tileset_path;
        }
        
        // Define Tileset Size
        u8 array_size = stage_layout_tiles_width * stage_layout_tiles_height;
        
        // Allocate Array Memory
        stage_layout_tiles_foreground = new u8[array_size];
        stage_layout_tiles_background = new u8[array_size];
        
        // Get Tileset Foreground Layer
        for (u32 i = 0; i < array_size; ++i)
        {
            stage_layout_tiles_foreground[i] = fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(u8));
        }
        
        // Get Background Layer Flag
        stage_draw_tiles_background = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Get Tileset Background Layer
        for (u32 i = 0; i < array_size; ++i)
        {
            stage_layout_tiles_background[i] = fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(u8));
        }
        
        /************************/
        /* Load Stage Collision */
        /************************/
        
        // Collision
        std::string collision_path;
        
        // Get Collision Path
        u8 stage_collision_path_length = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Write each char of length to string
        for (u8 i = 0; i <= stage_collision_path_length; ++i)
        {
            // Get Char and set Pointer
            char c = (char)fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(stage_collision_path_length));
            
            // Add char to string
            collision_path += c;
        }

        // Load Collision
        stage_load_collision(collision_path.c_str());
        
        /********************/
        /* Load Stage Sound */
        /********************/
        
        // Sound
        std::string sound_path;
        
        // Get Sound Path Length
        u8 stage_sound_path_length = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        // Write each char of length to string
        for (u8 i = 0; i <= stage_sound_path_length; ++i)
        {
            // Get Char and set Pointer
            char c = (char)fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(stage_sound_path_length));
            
            // Add char to string
            sound_path += c;
        }

        // Check if we're already playing the requested BGM!
        if (pSound != NULL)
        {
            if (strcmp(stage_sound_path.c_str(), sound_path.c_str()) != 0)
            {
                // Stop Sound
                pSound->sound_stop(SOUNDTYPE_BGM);
                
                // Play Sound
                pSound->sound_play_bgm(sound_path, true, true);
                
                // Set New Sound Path
                stage_sound_path = sound_path;
            }
        }
        else
        {
            // Set New Sound Path
            stage_sound_path = sound_path;
        }
        
        /****************************/
        /* Load Stage Ambient Light */
        /****************************/
        
        // Set Ambient Light
        stage_light_ambient = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        /********************/
        /* Load Stage Water */
        /********************/
        
        // Water Dimensions
        stage_water_x = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        
        stage_water_y = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        
        stage_water_w = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        
        stage_water_h = fileop_read_s32(mapfile);
        fileop_seekg_file(mapfile, sizeof(s32));
        
        // Water Design
        stage_water_alpha = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        stage_water_ripples_size = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        stage_water_godray = fileop_read_u8(mapfile);
        fileop_seekg_file(mapfile, sizeof(u8));
        
        /*********************/
        /* Read Stage Actors */
        /*********************/
        
        /* Actor File Paths */
        
        std::string path = "";
        
        /* Read Stage Enemies */
        
        // Define Struct
        EnemyMapData sEnemyMapData;
        
        // Read Enemies from file
        for (u8 i = 0; i < stage_enemies; ++i)
        {
            // Get Enemy File Path Length
            sEnemyMapData.stage_enemy_file_path_length[i] = fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(u8));
            
            // Get Enemy File
            
            // Write each char of length to string
            for (u8 j = 0; j <= sEnemyMapData.stage_enemy_file_path_length[i]; ++j)
            {
                // Get Char and set Pointer
                char c = (char)fileop_read_u8(mapfile);
                fileop_seekg_file(mapfile, sizeof(u8));
                
                // Add char to string
                path += c;
            }
            
            // Write File Path and reset temporary Path
            sEnemyMapData.stage_enemy_file_path[i] = path;
            path = "";
            
            // Get Enemy X
            sEnemyMapData.stage_enemy_x[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
            
            // Get Enemy Y
            sEnemyMapData.stage_enemy_y[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
        }

        /* Read Stage Objects */
        
        // Define Struct
        ObjectMapData sObjectMapData;
        
        // Read Objects from Map File
        for (u8 i = 0; i < stage_objects; ++i)
        {
            // Get Object File Path Length
            sObjectMapData.stage_object_file_path_length[i] = fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(u8));
            
            // Get Object File
            
            // Write each char of length to string
            for (u8 j = 0; j <= sObjectMapData.stage_object_file_path_length[i]; ++j)
            {
                // Get Char and set Pointer
                char c = (char)fileop_read_u8(mapfile);
                fileop_seekg_file(mapfile, sizeof(u8));
                
                // Add char to string
                path += c;
            }
            
            // Write File Path and reset temporary Path
            sObjectMapData.stage_object_file_path[i] = path;
            path = "";
            
            // Get Object X
            sObjectMapData.stage_object_x[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
            
            // Get Object Y
            sObjectMapData.stage_object_y[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
        }
        
        /* Read Stage Items */
        
        // Define Struct
        ItemMapData sItemMapData;
        
        // Read Items from Map File
        for (u8 i = 0; i < stage_items; ++i)
        {
            // Get Item File Path Length
            sItemMapData.stage_item_file_path_length[i] = fileop_read_u8(mapfile);
            fileop_seekg_file(mapfile, sizeof(u8));
            
            // Get Item File
            
            // Write each char of length to string
            for (u8 j = 0; j <= sItemMapData.stage_item_file_path_length[i]; ++j)
            {
                // Get Char and set Pointer
                char c = (char)fileop_read_u8(mapfile);
                fileop_seekg_file(mapfile, sizeof(u8));
                
                // Add char to string
                path += c;
            }
            
            // Write File Path and reset temporary Path
            sItemMapData.stage_item_file_path[i] = path;
            path = "";
            
            // Get Item X
            sItemMapData.stage_item_x[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
            
            // Get Item Y
            sItemMapData.stage_item_y[i] = fileop_read_s32(mapfile);
            fileop_seekg_file(mapfile, sizeof(s32));
        }
        
        /*********************/
        /* Load Stage Actors */
        /*********************/
        
        // Load Enemies
        if (stage_enemies > 0)
        {
            stage_load_enemies(pSpriteHandler, sEnemyMapData);
        }
        
        // Load Objects
        if (stage_objects > 0)
        {
            stage_load_objects(pSpriteHandler, sObjectMapData);
        }
        
        // Load Items
        if (stage_items > 0)
        {
            stage_load_items(sItemMapData);
        }
        
        /* Close Room File */
        
        mapfile.close();
        
        // Set File Pointer to 0 (Beginning of the File)
        fileop_seekg_resetpos();
        
        /* Set Room to loaded */
        
        stage_room_loaded = true;
    }
    else
    {
        // Flag as not loaded
        stage_room_loaded = false;
        
        // Force Kill
        exit(1);
    }
}

void Stage::stage_load_collision(const char* file)
{
    // Open Collision File
    FILE* fp = fopen(file, "rb");
    
    // Read Data
    if (!fp)
    {
        // Fill with 0
        memset(stage_collision, 0, (stage_width * stage_height) - 1);
        
    }
    else
    {
        // Define Variables
        u32 collision_data_size;

        // Get File Size
        fseek(fp, 0, SEEK_END);
        collision_data_size = ftell(fp);
        
        // Create Collision Array
        stage_collision = (u8*)malloc(collision_data_size + 1);//new u8[collision_data_size + 1];
        
        // Copy File Data
        fseek(fp, 0, SEEK_SET);
        fread(stage_collision, 1, collision_data_size, fp);
        
        // Close File
        fclose(fp);
    }
}

void Stage::stage_load_enemies(SpriteHandler* pSpriteHandler, EnemyMapData& sEnemyMapData)
{
    // Data
    EnemyFileStruct* pEnemyFileStruct[stage_enemies];
    std::vector<std::string> EnemyFilePath;
    u8 j = 0, k = 0;
    bool load_new_enemy_file = false;
    
    // Reset Data
    for (u8 i = 0; i <= stage_enemies - 1; ++i)
    {
        pEnemyFileStruct[i] = NULL;
        EnemyFilePath.push_back("NONE");
    }
    
    // Load Data
    while (j <= stage_enemies - 1)
    {
        // Create Enemy
        if (pEnemy[j] == NULL)
        {
            pEnemy[j] = new Enemy();
            
            // Load Enemy and increment Array Pos
            if (pEnemy[j] != NULL)
            {
                // Load Enemy File
                while (k <= stage_enemies - 1)
                {
                    // Check if we already loaded our Enemy File
                    if (strcmp(EnemyFilePath[k].c_str(), sEnemyMapData.stage_enemy_file_path[j].c_str()) == 0)
                    {
                        // Load Enemy
                        pEnemy[j]->enemy_load(pEnemyFileStruct[k], sEnemyMapData.stage_enemy_x[j], sEnemyMapData.stage_enemy_y[j]);
                        
                        // Exit Loop
                        break;
                    }
                    
                    // Increment Counter
                    ++k;
                    
                    // Check if we need to load a new enemy File
                    if (k > stage_enemies - 1)
                    {
                        // Load New File
                        load_new_enemy_file = true;
                    }
                }
                
                // Load new Enemy File
                if (load_new_enemy_file == true)
                {
                    // Reset Counter
                    k = 0;
                    
                    // Check already used Paths
                    while (strcmp(EnemyFilePath[k].c_str(), LINE_DO_NOT_USE) != 0 && k <= stage_enemies - 1)
                    {
                        ++k;
                    }
                    
                    // Load Enemy and matching Enemy File
                    if (k <= stage_enemies - 1)
                    {
                        // Load new File
                        pEnemyFileStruct[k] = fileop_load_enemy(sEnemyMapData.stage_enemy_file_path[j]);
                        
                        // Save File Path
                        EnemyFilePath[k] = sEnemyMapData.stage_enemy_file_path[j];
                        
                        // Load Enemy
                        pEnemy[j]->enemy_load(pEnemyFileStruct[k], sEnemyMapData.stage_enemy_x[j], sEnemyMapData.stage_enemy_y[j]);
                    }
                    
                    // Reset Load File and Counter
                    load_new_enemy_file = false;
                    k = 0;
                }
                
                // Load Sprite
                if (strcmp(pEnemy[j]->enemy_get_sprite_path().c_str(), LINE_DO_NOT_USE) != 0)
                {
                    pSpriteHandler->spritehandler_load_sprite(pEnemy[j]->enemy_get_sprite_path(),
                                                              pEnemy[j]->enemy_get_max_spriteline(),
                                                              pEnemy[j]->enemy_get_sprite_width(),
                                                              pEnemy[j]->enemy_get_sprite_height(),
                                                              pEnemy[j]->enemy_get_sprite_frame_speed());
                }
                
                ++j;
            }
        }
        
        // We reached the Enemy Count, exit loop
        if (j == stage_enemies)
            break;
    }
    
    // Delete all Enemy Files
    for (u8 i = 0; i <= stage_enemies - 1; ++i)
    {
        // Check if we used the Slot
        if (pEnemyFileStruct[i] != NULL)
        {
            // Delete Enemy File
            delete pEnemyFileStruct[i];
        }
    }
}

void Stage::stage_load_items(ItemMapData& sItemMapData)
{
    // Data
    u8 j = 0;
    
    // Load Data
    while (j <= stage_items - 1)
    {
        // Create Item
        if (pItem[j] == NULL)
        {
            // Allocate Memory and Initialize Item Class
            pItem[j] = new Item();
            
            // Load Item
            if (pItem[j] != NULL)
            {
                // Load Item
                pItem[j]->item_load(sItemMapData.stage_item_file_path[j], sItemMapData.stage_item_x[j], sItemMapData.stage_item_y[j]);
            }
            
            // Increment Array Position Counter
            ++j;
        }
        
        // We reached the Object Count, exit loop
        if (j == stage_items)
            break;
    }
}

void Stage::stage_load_objects(SpriteHandler* pSpriteHandler, ObjectMapData& sObjectMapData)
{
    // Data
    ObjectFileStruct* pObjectFileStruct[stage_objects];
    std::vector<std::string> ObjectFilePath;
    u8 j = 0, k = 0;
    bool load_new_object_file = false;
    
    // Reset Data
    for (u8 i = 0; i <= stage_objects - 1; ++i)
    {
        pObjectFileStruct[i] = NULL;
        ObjectFilePath.push_back("NONE");
    }
    
    // Load Data
    while (j <= stage_objects - 1)
    {
        // Create Object
        if (pObject[j] == NULL)
        {
            // Allocate Memory and Initialize Object Class
            pObject[j] = new Object();
            
            // Load Object and increment Array Pos
            if (pObject[j] != NULL)
            {
                // Load Object File
                while (k <= stage_objects - 1)
                {
                    // Check if we already loaded our Object File
                    if (strcmp(ObjectFilePath[k].c_str(), sObjectMapData.stage_object_file_path[j].c_str()) == 0)
                    {
                        // Load Object
                        pObject[j]->object_load(pObjectFileStruct[k], sObjectMapData.stage_object_x[j], sObjectMapData.stage_object_y[j]);
                        
                        // Exit Loop
                        break;
                    }
                    
                    // Increment Counter
                    ++k;
                    
                    // Check if we need to load a new enemy File
                    if (k > stage_objects - 1)
                    {
                        // Load New File
                        load_new_object_file = true;
                    }
                }
                
                // Load new Enemy File
                if (load_new_object_file == true)
                {
                    // Reset Counter
                    k = 0;
                    
                    // Check already used Paths
                    while (strcmp(ObjectFilePath[k].c_str(), LINE_DO_NOT_USE) != 0 && k <= stage_objects - 1)
                    {
                        ++k;
                    }
                    
                    // Load Object and matching Object File
                    if (k <= stage_objects - 1)
                    {
                        // Create Struct
                        pObjectFileStruct[k] = new ObjectFileStruct();
                        
                        // Load new File
                        pObjectFileStruct[k] = fileop_load_object(sObjectMapData.stage_object_file_path[j]);
                        
                        // Save File Path
                        ObjectFilePath[k] = sObjectMapData.stage_object_file_path[j];
                        
                        // Load Object
                        pObject[j]->object_load(pObjectFileStruct[k], sObjectMapData.stage_object_x[j], sObjectMapData.stage_object_y[j]);
                    }
                    
                    // Reset Load File and Counter
                    load_new_object_file = false;
                    k = 0;
                }
                
                // Load Sprite
                if (strcmp(pObject[j]->object_get_sprite_path().c_str(), LINE_DO_NOT_USE) != 0)
                {
                    pSpriteHandler->spritehandler_load_sprite(pObject[j]->object_get_sprite_path(),
                                                              pObject[j]->object_get_max_spriteline(),
                                                              pObject[j]->object_get_sprite_width(),
                                                              pObject[j]->object_get_sprite_height(),
                                                              pObject[j]->object_get_sprite_frame_speed());
                }
                
                ++j;
            }
        }
        
        // We reached the Object Count, exit loop
        if (j == stage_objects)
            break;
    }
    
    // Delete all Object Files
    for (u8 i = 0; i <= stage_objects - 1; ++i)
    {
        // Check if we used the Slot
        if (pObjectFileStruct[i] != NULL)
        {
            // Delete Enemy File
            delete pObjectFileStruct[i];
        }
    }
}

void Stage::stage_create_item(std::string item_file, s32 item_x, s32 item_y)
{
    // Check if we got a valid Item!
    if (strcmp(item_file.c_str(), LINE_DO_NOT_USE) == 0)
    {
        // Return on Loot unsued
        return;
    }
    
    // Create in new Slot bool
    bool create_new_slot = true;
    
    // Check old slots for unused ones
    if (stage_items > 0)
    {
        // Check for old Itemslots which are free
        for (u8 i = 0; i <= stage_items - 1; ++i)
        {
            // Create Item
            if (pItem[i] == NULL)
            {
                pItem[i] = new Item();
                
                // Load Item and increment Array Pos
                if (pItem[i] != NULL)
                {
                    pItem[i]->item_load(item_file, item_x, item_y);
                    create_new_slot = false;
                }
                
                // Break Loop
                break;
            }
        }
    }
    
    // Create Item in new Slot
    if (create_new_slot == true && stage_items < STAGE_MAX_ITEMS && pItem[stage_items] == NULL)
    {
        pItem[stage_items] = new Item();
        
        // Load Item and increment Array Pos
        if (pItem[stage_items] != NULL)
        {
            pItem[stage_items]->item_load(item_file, item_x, item_y);
        }
        
        // Increment Item Counter
        ++stage_items;
    }
}

/*************************************************************************************************************************/
/*                                                   Stage Deleting                                                      */
/*************************************************************************************************************************/

void Stage::stage_delete()
{
    /* Delete Stage */

    /* Check for Loaded Info */
    
    if (stage_room_loaded == false)
    {
        return;
    }

    /* Delete Spawned Actors */
    
    stage_delete_actors();
    
    /* Delete Environment */
    
    stage_delete_environment();

    /* Reset Variables */

    stage_reset_variables();
}

void Stage::stage_delete_environment()
{
    if (stage_background != NULL)
    {
        sf2d_free_texture(stage_background);
        stage_background = NULL;
    }
    
    if (stage_collision != NULL)
    {
        delete[] stage_collision;
        stage_collision = NULL;
    }
    
    if (stage_tileset != NULL)
    {
        sf2d_free_texture(stage_tileset);
        stage_tileset = NULL;
    }
    
    if (stage_layout_tiles_foreground != NULL)
    {
        delete[] stage_layout_tiles_foreground;
        stage_layout_tiles_foreground = NULL;
    }
    
    if (stage_layout_tiles_background != NULL)
    {
        delete[] stage_layout_tiles_background;
        stage_layout_tiles_background = NULL;
    }
}

void Stage::stage_delete_environment(std::string new_background_path, std::string new_tileset_path)
{
    if (stage_background != NULL && strcmp(stage_background_path.c_str(), new_background_path.c_str()) != 0)
    {
        sf2d_free_texture(stage_background);
        stage_background = NULL;
    }
    
    if (stage_collision != NULL)
    {
        delete[] stage_collision;
        stage_collision = NULL;
    }
    
    if (stage_tileset != NULL && strcmp(stage_tileset_path.c_str(), new_tileset_path.c_str()) != 0)
    {
        sf2d_free_texture(stage_tileset);
        stage_tileset = NULL;
    }
    
    if (stage_layout_tiles_foreground != NULL)
    {
        delete[] stage_layout_tiles_foreground;
        stage_layout_tiles_foreground = NULL;
    }
    
    if (stage_layout_tiles_background != NULL)
    {
        delete[] stage_layout_tiles_background;
        stage_layout_tiles_background = NULL;
    }
}

void Stage::stage_reset_variables()
{
    /* Reset Stage Variables */
    
    stage_id = 0;
    stage_phase = 0;
    stage_width = 0;
    stage_height = 0;
    stage_minimap_id = 0;
    stage_light_ambient = 0;
    stage_water_type = 0;
    stage_water_x = 0;
    stage_water_y = 0;
    stage_water_w = 0;
    stage_water_h = 0;
    stage_water_alpha = 0;
    stage_water_ripples_size = 0;
    stage_water_godray = false;
    stage_godray_x = 0;
    stage_godray_y = 0;
}

void Stage::stage_delete_actors()
{
    /* Delete Spawned Actors */
    
    // Delete Enemies
    if (stage_enemies > 0)
    {
        for (u8 i = 0; i <= stage_enemies - 1; ++i)
        {
            if (pEnemy[i] != NULL)
                delete pEnemy[i];
            
            pEnemy[i] = NULL;
        }
        
        stage_enemies = 0;
    }
    
    // Delete Items
    if (stage_items > 0)
    {
        for (u8 i = 0; i <= stage_items - 1; ++i)
        {
            if (pItem[i] != NULL)
                delete pItem[i];
            
            pItem[i] = NULL;
        }
        
        stage_items = 0;
    }
    
    // Delete Objects
    if (stage_objects > 0)
    {
        for (u8 i = 0; i <= stage_objects - 1; ++i)
        {
            if (pObject[i] != NULL)
                delete pObject[i];
            
            pObject[i] = NULL;
        }
        
        stage_objects = 0;
    }
}

/*************************************************************************************************************************/
/*                                                   Stage Info Returns                                                  */
/*************************************************************************************************************************/

u8 Stage::stage_get_layer_tile(bool is_background_layer, u32 array_pos)
{
    // Check if we have a tileset loaded
    if (stage_tileset == NULL)
    {
        return 0;
    }
    
    // We decide between back- and foreground layers
    if (is_background_layer == false)
    {
        // Check if we have the requested tile layer loaded
        if (stage_layout_tiles_foreground != NULL && array_pos < (stage_layout_tiles_width * stage_layout_tiles_height))
        {
            return stage_layout_tiles_foreground[array_pos];
        }
        else
        {
            return 0;
        }
    }
    else
    {
        // Check for loaded layout
        if (stage_layout_tiles_background != NULL && stage_draw_tiles_background == true && array_pos < (stage_layout_tiles_width * stage_layout_tiles_height))
        {
            return stage_layout_tiles_background[array_pos];
        }
        else
        {
            return 0;
        }
    }
}

Enemy* Stage::stage_get_enemy(u8 pos)
{
    if (pos > stage_enemies || pos > STAGE_MAX_ENEMIES)
        return NULL;
    
    return pEnemy[pos];
}

Item* Stage::stage_get_item(u8 pos)
{
    if (pos > stage_items || pos > STAGE_MAX_ITEMS)
        return NULL;
    
    return pItem[pos];
}

Object* Stage::stage_get_object(u8 pos)
{
    if (pos > stage_objects || pos > STAGE_MAX_OBJECTS)
        return NULL;
    
    return pObject[pos];
}

void Stage::stage_delete_object(u8 pos)
{
    if (pos <= stage_objects - 1)
    {
        // Delete Object if used
        if (pObject[pos] != NULL)
        {
            delete pObject[pos];
            pObject[pos] = NULL;
        }
    }
}

void Stage::stage_delete_enemy(u8 pos)
{
    if (pos <= stage_enemies - 1)
    {
        // Delete Object if used
        if (pEnemy[pos] != NULL)
        {
            delete pEnemy[pos];
            pEnemy[pos] = NULL;
        }
    }
}

void Stage::stage_delete_item(u8 pos)
{
    if (pos <= stage_items - 1)
    {
        // Delete Object if used
        if (pItem[pos] != NULL)
        {
            // Delete Item and set to NULL
            delete pItem[pos];
            pItem[pos] = NULL;
        }
    }
}

bool Stage::stage_is_in_water(s32 actor_x, s32 actor_y, s32 actor_h)
{
    if (actor_x >= stage_water_x && actor_x <= stage_water_x + stage_water_w &&
        actor_y + (actor_h / 2) < stage_water_y && actor_y + (actor_h / 2) >= stage_water_y - stage_water_h)
    {
        return true;
    }
    else
    {
        return false;
    }
}

bool Stage::stage_is_in_water_steps(s32 actor_x, s32 actor_y, s32 actor_h)
{
    if (actor_x >= stage_water_x && actor_x <= stage_water_x + stage_water_w &&
        actor_y + 1 < stage_water_y && actor_y + 1 >= stage_water_y - stage_water_h)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*************************************************************************************************************************/
/*                                                   Stage Collision                                                     */
/*************************************************************************************************************************/

s32 Stage::stage_get_collision_x(u8 collision, s32 object_x, s32 object_distance, s32 object_w)
{
    // We expect X to be in the middle of the Sprite shown on screen!
    s32 return_value = 0;

    if (collision == COLLISION_LEFT)
    {
        if (stage_width < (object_x))
            return_value = stage_width;
        else if (0 > (object_x))
            return_value = 0;
        else
            return_value = object_x - 1;
    }
    else if (collision == COLLISION_RIGHT)
    {
        if (stage_width < (object_x))
            return_value = stage_width;
        else if (0 > (object_x))
            return_value = 0;
        else
            return_value = object_x + 1;
    }
    else
    {
        if (stage_width < object_x)
            return_value = stage_width;
        else if (0 > object_x)
            return_value = 0;
        else
            return_value = object_x;
    }
    
    // Return Value
    return return_value;
}

s32 Stage::stage_get_collision_y(u8 collision, s32 object_y, s32 object_distance, s32 object_height)
{
    // We expect Y to be at the bottom of the Sprite shown on screen!
    s32 return_value = 0;

    if (collision == COLLISION_UP)
    {
        if (stage_height < (object_y + object_height + 1))
        {
            return_value = stage_height - 1; // 0
        }
        else if (0 > (object_y + object_height + 1))
        {
            return_value = 0; // stage_height - 1
        }
        else if (stage_height == object_y)
        {
            return 0; //stage_height - 1;
        }
        else
        {
            return_value = stage_height - (object_y + object_height + 1);
        }
    }
    else if (collision == COLLISION_DOWN)
    {
        if (stage_height < (object_y - 1))
        {
            return_value = stage_height - 1; //0;
        }
        else if (0 > (object_y - 1))
        {
            return_value = 0; //stage_height;
        }
        else if (stage_height == object_y)
        {
            return 0; //stage_height - 1;
        }
        else
        {
            return_value = stage_height - (object_y - 1);
        }
    }
    else
    {
        if (stage_height < object_y + object_height)
        {
            return_value = stage_height; //0;
        }
        else if (0 > object_y + object_height)
        {
            return_value = 0; //stage_height;
        }
        else if (stage_height == object_y)
        {
            return stage_height - 1;
        }
        else
        {
            return_value = stage_height - (object_y + object_height); // Collision for Left and Right
        }
    }
    
    // Return Value
    return return_value;
}

u8 Stage::stage_check_collision(s32 object_x, s32 object_y, s32 object_width, s32 object_height)
{
    // Set Variables
    u8 stage_actor_collision = COLLISION_NONE;
    
    // Return on Actor X / Y greater or smaller than stage bounds
    if (object_x > stage_width || object_x < 0 || object_y + object_height > stage_height || object_y < 0)
    {
        return stage_actor_collision;
    }

    /* Check for Down Collision */
    
    // Get Down Pixel
    if (stage_collision[stage_get_collision_x(COLLISION_DOWN, object_x, 0, object_width) + (stage_get_collision_y(COLLISION_DOWN, object_y, 0, object_height) * stage_width)] != 0)
    {
        // Add Down Collision
        stage_actor_collision += COLLISION_DOWN;
    }
    
    /* Check for Up Collision */
    
    // Get Up Pixel
    if (stage_collision[stage_get_collision_x(COLLISION_UP, object_x, 0, object_width) + (stage_get_collision_y(COLLISION_UP, object_y, 0, object_height) * stage_width)] != 0)
    {
        // Add Up Collision
        stage_actor_collision += COLLISION_UP;
    }
    
    /* Check for Left Collision */
    
    // Get Left Pixel
    if (stage_collision[stage_get_collision_x(COLLISION_LEFT, object_x, 0, object_width) + (stage_get_collision_y(COLLISION_LEFT, object_y, 0, object_height) * stage_width)] != 0)
    {
        // Add Left Collision
        stage_actor_collision += COLLISION_LEFT;
    }
    
    /* Check for Right Collision */
    
    // Get Pixel
    if (stage_collision[stage_get_collision_x(COLLISION_RIGHT, object_x, 0, object_width) + (stage_get_collision_y(COLLISION_RIGHT, object_y, 0, object_height) * stage_width)] != 0)
    {
        // Add Right Collision
        stage_actor_collision += COLLISION_RIGHT;
    }
    
    /* Return Collision Result */
    
    return stage_actor_collision;
}

u8 Stage::stage_check_collision_distance(s32 object_x, s32 object_y, s32 object_width, s32 object_height, u8 distance)
{
    // Set Variables
    u8 stage_actor_collision = COLLISION_NONE;
    
    // Return on Actor X / Y greater or smaller than stage bounds
    if (object_x > stage_width || object_x < 0 || object_y + object_height > stage_height || object_y < 0)
    {
        return stage_actor_collision;
    }
    
    /* Check for Down Collision */
    
    for (u8 i = distance; i > 0; --i) // 0 is ignored, it is our Y Position!
    {
        // Get Down Pixel
        if (stage_collision[stage_get_collision_x(COLLISION_DOWN, object_x, distance, object_width) + (stage_get_collision_y(COLLISION_DOWN, object_y + i, distance, object_height) * stage_width)] != 0)
        {
            // Set Down Collision
            stage_actor_collision += COLLISION_DOWN;
            
            // Break Look
            break;
        }
    }
    
    
    /* Check for Up Collision */
    
    for (u8 i = distance; i > 0; --i)
    {
        // Get Up Pixel
        if (stage_collision[stage_get_collision_x(COLLISION_UP, object_x, distance, object_width) + (stage_get_collision_y(COLLISION_UP, object_y - i, distance, object_height) * stage_width)] != 0)
        {
            // Set Up Collision
            stage_actor_collision += COLLISION_UP;
            
            // Break Loop
            break;
        }
    }
    
    
    /* Check for Left Collision */
    
    for (u8 i = distance; i > 0; --i)
    {
        for (s32 j = 0; j <= object_height; ++j)
        {
            // Get Pixel
            if (stage_collision[stage_get_collision_x(COLLISION_LEFT, object_x - i, distance, object_width) + (stage_get_collision_y(COLLISION_LEFT, object_y, distance, j) * stage_width)] != 0)
            {
                // Add Left Collision
                stage_actor_collision += COLLISION_LEFT;
                
                // Break Loop
                break;
            }
            
            // Check for Collision Result
            if (stage_actor_collision & COLLISION_LEFT)
            {
                break;
            }
        }
    }
    
    
    /* Check for Right Collision */
    
    for (u8 i = distance; i > 0; --i)
    {
        for (s32 j = 0; j <= object_height; ++j)
        {
            // Get Pixel
            if (stage_collision[stage_get_collision_x(COLLISION_RIGHT, object_x + i, distance, object_width) + (stage_get_collision_y(COLLISION_RIGHT, object_y, distance, j) * stage_width)] != 0)
            {
                // Add Right Collision
                stage_actor_collision += COLLISION_RIGHT;
                
                // Break Loop
                break;
            }
        
            // Check for Collision Result
            if (stage_actor_collision & COLLISION_RIGHT)
            {
                break;
            }
        }
    }
    
    /* Return Collision Result */
    
    return stage_actor_collision;
}

s32 Stage::stage_check_specific_collision_distance(s32 object_x, s32 object_y, s32 object_width, s32 object_height, u8 distance, u8 collision)
{
    /* Check for Down Collision */
    
    if (collision & COLLISION_DOWN)
    {
        // Return on Actor Y < 0
        if (object_y < 0)
        {
            return object_y - 1;
        }
        
        for (u8 i = 1; i <= distance; ++i) // 0 is ignored, it is our Y Position!
        {
            // Get Down Pixel
            if (stage_collision[stage_get_collision_x(COLLISION_DOWN, object_x, distance, object_width) + (stage_get_collision_y(COLLISION_DOWN, object_y - i, distance, object_height) * stage_width)] != 0)
            {
                // Return Collision Coordinates
                return object_y - i; // - i because our Object Y is normally Y 0 = Bottom of the Screen
            }
        }
        
        // Return Default Error. -1 Signals no Collision in the selected Distance
        return object_y - distance;
    }
    
    /* Check for Up Collision */
    
    if (collision & COLLISION_UP)
    {
        // Return on Actor Y + Actor Height > Stage Height
        if (object_y + object_height > stage_height)
        {
            return object_y + 1;
        }
        
        for (s8 i = distance; i >= 0; --i)
        {
            // Get Up Pixel
            if (stage_collision[stage_get_collision_x(COLLISION_UP, object_x, distance, object_width) + (stage_get_collision_y(COLLISION_UP, object_y + i, distance, object_height) * stage_width)] != 0)
            {
                // Return Collision Coordinates
                return object_y + i; // +1 because Y MAX = Screen Top
            }
        }
        
        // Return Default Error
        return object_y + distance + 1;
    }
    
    /* Check for Left Collision */
    
    if (collision & COLLISION_LEFT)
    {
        // Return on Actor X < 0
        if (object_x < 0)
        {
            return object_x;
        }
        
        for (u8 i = 1; i <= distance; ++i)
        {
            for (s32 j = 0; j <= object_height; ++j)
            {
                // Get Pixel
                if (stage_collision[stage_get_collision_x(COLLISION_LEFT, object_x - i, distance, object_width) + (stage_get_collision_y(COLLISION_LEFT, object_y, distance, j) * stage_width)] != 0)
                {
                    // Return Collision Result
                    return object_x - (i - 1);
                }
            }
        }
        
        // Return Default Error
        return object_x - distance;
    }
    
    /* Check for Right Collision */
    
    if (collision & COLLISION_RIGHT)
    {
        // Return on Actor X > Stage Width
        if (object_x > stage_width)
        {
            return object_x;
        }
        
        for (u8 i = 1; i <= distance; ++i)
        {
            for (s32 j = 0; j <= object_height; ++j)
            {
                // Get Pixel
                if (stage_collision[stage_get_collision_x(COLLISION_RIGHT, object_x + i, distance, object_width) + (stage_get_collision_y(COLLISION_RIGHT, object_y, distance, j) * stage_width)] != 0)
                {
                    // Return Collision Result
                    return object_x + (i - 1);
                }
            }
        }
        
        // Return Default Error
        return object_x + distance;
    }
    
    /* Return Default Distance */
    
    return -1;
}
