//
//  stage.hpp
//  3ds_metroid
//
//  Created by jbr373 on 16/01/16.
//

#ifndef stage_h
#define stage_h

#include <stdio.h>
#include <stdlib.h>
#include <3ds.h>
#include <string>

#include "image.h"
#include "enemy.h"
#include "item.h"
#include "object.h"
#include "sprite_handler.h"
#include "file_operations.h"

#define STAGE_MAX_ENEMIES           15 // 0 - 15 => 16 Enemies
#define STAGE_MAX_ITEMS             15 // 0 - 19 => 20 Items (Enemy Drops + Placed Items)
#define STAGE_MAX_OBJECTS           29

enum StageLoopType
{
    STAGE_LAYOUT_SINGE = 0, // No Second Part
    STAGE_LAYOUT_LEFTRIGHT = 1,
    STAGE_LAYOUT_UPDOWN = 2
};

enum StageRoomChange
{
    STAGE_ROOMCHANGE_NOCHANGE = 0,
    
    STAGE_ROOMCHANGE_FADEOUT = 1,
    STAGE_ROOMCHANGE_LOAD = 2,
    STAGE_ROOMCHANGE_FADEIN = 3,
};

struct EnemyMapData;
struct ObjectMapData;
struct ItemMapData;

class Stage : private FileOperations
{
public:
    Stage();
    virtual ~Stage();
    
    // Stage Loading
    void stage_load(Sound* pSound, SpriteHandler* pSpriteHandler, u16 room_id);
    
    // Stage Deletion
    void stage_delete();
    void stage_delete_object(u8 pos);
    void stage_delete_enemy(u8 pos);
    void stage_delete_item(u8 pos);
    
    // Room Interaction
    void stage_create_item(std::string item_file, s32 item_x, s32 item_y);
    
    u8 stage_check_objectpos(s32 x, s32 y);
    
    u8 stage_check_collision(s32 object_x, s32 object_y, s32 object_width, s32 object_height);
    u8 stage_check_collision_distance(s32 object_x, s32 object_y, s32 object_width, s32 object_height, u8 distance);
    s32 stage_check_specific_collision_distance(s32 object_x, s32 object_y, s32 object_width, s32 object_height, u8 distance, u8 collision);
    
    bool stage_is_in_water(s32 actor_x, s32 actor_y, s32 actor_h);
    bool stage_is_in_water_steps(s32 actor_x, s32 actor_y, s32 actor_h);
    s32 stage_get_next_y(s32 object_x, s32 object_y, s32 object_width, s32 object_height, u8 object_facing);
    
    // Getters
    Enemy* stage_get_enemy(u8 pos);
    Object* stage_get_object(u8 pos);
    Item* stage_get_item(u8 pos);
    
    u8 stage_get_enemy_count() { return stage_enemies; };
    u8 stage_get_object_count() { return stage_objects; };
    u8 stage_get_items_count() { return stage_items; };
    u8 stage_get_stage_phase() { return stage_phase; };
    u8 stage_get_light() { return stage_light_ambient; };
    u8 stage_get_tile_layout_width() { return stage_layout_tiles_width; };
    u8 stage_get_tile_layout_height() { return stage_layout_tiles_height; };
    u8 stage_get_layer_tile(bool is_background_layer, u32 array_pos);
    u8 stage_get_room_change() { return stage_room_change; };
    u8 stage_get_water_type() { return stage_water_type; };
    
    u16 stage_get_id() { return stage_id; };
    
    u32 stage_get_minimap_id() { return stage_minimap_id; };
    u32 stage_get_tile_count() { return stage_layout_tiles_width * stage_layout_tiles_height; };
    
    s32 stage_get_width() { return stage_width; };
    s32 stage_get_height() { return stage_height; };
    s32 stage_get_water_x() { return stage_water_x; };
    s32 stage_get_water_y() { return stage_water_y; };
    s32 stage_get_water_w() { return stage_water_w; };
    s32 stage_get_water_h() { return stage_water_h; };

    bool stage_get_loaded() { return stage_room_loaded; };
    bool stage_get_background_tileset_draw_flag() { return stage_draw_tiles_background; };
    
    std::string stage_get_background_path() { return stage_background_path; };
    std::string stage_get_tileset_path() { return stage_tileset_path; };
    std::string stage_get_sound_path() { return stage_sound_path; };
    
    sf2d_texture* stage_get_background() { return stage_background; };
    sf2d_texture* stage_get_tileset() { return stage_tileset; };
    
    // Setters
    void stage_set_object_count(u8 objects) {stage_objects = objects; };
    void stage_set_room_change(u8 state) { stage_room_change = state; };
    void stage_set_sound_path(std::string path) { stage_sound_path = path; };
    void stage_set_background_path(std::string path) { stage_background_path = path; };
    void stage_set_tileset_path(std::string path) { stage_tileset_path = path; };

    
private:
    // Stage Loading
    void stage_load_enemies(SpriteHandler* pSpriteHandler, EnemyMapData& sEnemyMapData);
    void stage_load_items(ItemMapData& sItemMapData);
    void stage_load_objects(SpriteHandler* pSpriteHandler, ObjectMapData& sObjectMapData);
    void stage_load_collision(const char* file);
    
    // Stage Deletion
    void stage_delete_environment();
    void stage_delete_environment(std::string new_background_path, std::string new_tileset_path);
    void stage_reset_variables();
    void stage_delete_actors();
    
    // Coordinates
    s32 stage_get_collision_x(u8 collision, s32 object_x, s32 object_distance, s32 object_width);
    s32 stage_get_collision_y(u8 collision, s32 object_y, s32 object_distance, s32 object_height);

    // Actors
    Enemy* pEnemy[STAGE_MAX_ENEMIES];
    Item* pItem[STAGE_MAX_ITEMS];
    Object* pObject[STAGE_MAX_OBJECTS];
    
    // Loading & Animation
    u8 stage_room_change;
    
    /* Stage Room Data */
    
    bool stage_room_loaded;
    
    u16 stage_id;
    u8 stage_phase;
    u8 stage_enemies;
    u8 stage_items;
    u8 stage_objects;
    
    s32 stage_width;
    s32 stage_height;
    
    u32 stage_minimap_id;
    
    sf2d_texture* stage_background;
    sf2d_texture* stage_tileset;
    
    std::string stage_background_path;
    std::string stage_tileset_path;
    std::string stage_sound_path;
    
    u8 stage_layout_tiles_width; // Tiles in a line from Left to Right - 0 = no Tiles
    u8 stage_layout_tiles_height; // Tiles in a line from up to down - 0 = no Tiles
    u8* stage_layout_tiles_foreground;
    u8* stage_layout_tiles_background;
    bool stage_draw_tiles_background;
    u8* stage_collision;
    
    u8 stage_light_ambient;
    
    u8 stage_water_type;
    s32 stage_water_x;
    s32 stage_water_y;
    s32 stage_water_w;
    s32 stage_water_h;
    u8 stage_water_alpha;
    u8 stage_water_ripples_size;
    
    bool stage_water_godray;
    s32 stage_godray_x;
    s32 stage_godray_y;
};

#endif /* stage_h */
